#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from heat.common import exception
from heat.common.i18n import _
from heat.engine.clients import progress
from heat.engine import resource
from heat.engine import rsrc_defn


class BaseVolume(resource.Resource):
    """Base Volume Manager."""

    default_client_name = 'cinder'

    def _create_arguments(self):
        return {}

    def handle_create(self):
        backup_id = self.properties.get(self.BACKUP_ID)
        cinder = self.client()
        if backup_id is not None:
            vol_id = cinder.restores.restore(backup_id).volume_id
            vol = cinder.volumes.get(vol_id)
            kwargs = self._fetch_name_and_description()
            cinder.volumes.update(vol_id, **kwargs)
        else:
            kwargs = self._create_arguments()
            kwargs.update(self._fetch_name_and_description())
            vol = cinder.volumes.create(**kwargs)
        self.resource_id_set(vol.id)

        return vol.id

    def check_create_complete(self, vol_id):
        vol = self.client().volumes.get(vol_id)

        if vol.status == 'available':
            return True
        if vol.status in self._volume_creating_status:
            return False
        if vol.status == 'error':
            raise exception.ResourceInError(
                resource_status=vol.status)
        else:
            raise exception.ResourceUnknownStatus(
                resource_status=vol.status,
                result=_('Volume create failed'))

    def _name(self):
        return self.physical_resource_name()

    def _description(self):
        return self.physical_resource_name()

    def _fetch_name_and_description(self, name=None,
                                    description=None):
        return {'name': name or self._name(),
                'description': description or self._description()}

    def handle_check(self):
        vol = self.client().volumes.get(self.resource_id)
        statuses = ['available', 'in-use']
        checks = [
            {'attr': 'status', 'expected': statuses, 'current': vol.status},
        ]
        self._verify_check_conditions(checks)

    def handle_snapshot_delete(self, state):
        backup = state not in ((self.CREATE, self.FAILED),
                               (self.UPDATE, self.FAILED))
        prg = progress.VolumeDeleteProgress()
        prg.backup['called'] = not backup
        prg.backup['complete'] = not backup
        return prg

    def handle_delete(self):
        if self.resource_id is None:
            return progress.VolumeDeleteProgress(True)
        prg = progress.VolumeDeleteProgress()
        prg.backup['called'] = True
        prg.backup['complete'] = True
        return prg

    def _create_backup(self):
        backup = self.client().backups.create(self.resource_id)
        return backup.id

    def _check_create_backup_complete(self, prg):
        backup = self.client().backups.get(prg.backup_id)
        if backup.status == 'creating':
            return False
        if backup.status == 'available':
            return True
        else:
            raise exception.ResourceUnknownStatus(
                resource_status=backup.status,
                result=_('Volume backup failed'))

    def _delete_volume(self):
        """Call the volume delete API.

        Returns False if further checking of volume status is required,
        True otherwise.
        """
        try:
            cinder = self.client()
            vol = cinder.volumes.get(self.resource_id)
            if vol.status == 'in-use':
                raise exception.Error(_('Volume in use'))
            # if the volume is already in deleting status,
            # just wait for the deletion to complete
            if vol.status != 'deleting':
                cinder.volumes.delete(self.resource_id)
            return False
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
            return True

    def check_delete_complete(self, prg):
        if not prg.backup['called']:
            prg.backup_id = self._create_backup()
            prg.backup['called'] = True
            return False

        if not prg.backup['complete']:
            prg.backup['complete'] = self._check_create_backup_complete(prg)
            return False

        if not prg.delete['called']:
            prg.delete['complete'] = self._delete_volume()
            prg.delete['called'] = True
            return False

        if not prg.delete['complete']:
            try:
                vol = self.client().volumes.get(self.resource_id)
            except Exception as ex:
                self.client_plugin().ignore_not_found(ex)
                prg.delete['complete'] = True
                return True
            if vol.status.lower() == 'error_deleting':
                raise exception.ResourceInError(status_reason='delete',
                                                resource_status=vol.status)
            else:
                return False
        return True

    @classmethod
    def validate_deletion_policy(cls, policy):
        res = super(BaseVolume, cls).validate_deletion_policy(policy)
        if res:
            return res
        if (policy == rsrc_defn.ResourceDefinition.SNAPSHOT and
                not cfg.CONF.volumes.backups_enabled):
            msg = _('"%s" deletion policy not supported - '
                    'volume backup service is not enabled.') % policy
            raise exception.StackValidationFailed(message=msg)


class BaseVolumeAttachment(resource.Resource):
    """Base Volume Attachment Manager."""

    default_client_name = 'cinder'

    def handle_create(self):
        server_id = self.properties[self.INSTANCE_ID]
        volume_id = self.properties[self.VOLUME_ID]
        dev = (self.properties[self.DEVICE] if self.properties[self.DEVICE]
               else None)

        attach_id = self.client_plugin('nova').attach_volume(
            server_id, volume_id, dev)

        self.resource_id_set(attach_id)

        return volume_id

    def check_create_complete(self, volume_id):
        return self.client_plugin().check_attach_volume_complete(volume_id)

    def handle_delete(self):
        prg = None

        if self.resource_id:
            server_id = self.properties[self.INSTANCE_ID]
            vol_id = self.properties[self.VOLUME_ID]
            prg = progress.VolumeDetachProgress(server_id, vol_id,
                                                self.resource_id)
            prg.called = self.client_plugin('nova').detach_volume(
                server_id, self.resource_id)

        return prg

    def check_delete_complete(self, prg):
        if prg is None:
            return True

        if not prg.called:
            prg.called = self.client_plugin('nova').detach_volume(
                prg.srv_id, self.resource_id)
            return False
        if not prg.cinder_complete:
            prg.cinder_complete = self.client_plugin(
            ).check_detach_volume_complete(prg.vol_id, prg.srv_id)
            return False
        if not prg.nova_complete:
            prg.nova_complete = self.client_plugin(
                'nova').check_detach_volume_complete(prg.srv_id,
                                                     prg.attach_id)
            return prg.nova_complete
        return True



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import fnmatch
import glob
import itertools
import os.path
import re
import weakref

from oslo_config import cfg
from oslo_log import log

from heat.common import environment_format as env_fmt
from heat.common import exception
from heat.common.i18n import _
from heat.common import policy
from heat.engine import support

LOG = log.getLogger(__name__)


HOOK_TYPES = (
    HOOK_PRE_CREATE, HOOK_PRE_UPDATE, HOOK_PRE_DELETE, HOOK_POST_CREATE,
    HOOK_POST_UPDATE, HOOK_POST_DELETE
) = (
    'pre-create', 'pre-update', 'pre-delete', 'post-create',
    'post-update', 'post-delete'
)

RESTRICTED_ACTIONS = (UPDATE, REPLACE) = ('update', 'replace')


def valid_hook_type(hook):
    return hook in HOOK_TYPES


def valid_restricted_actions(action):
    return action in RESTRICTED_ACTIONS


def is_hook_definition(key, value):
    is_valid_hook = False
    if key == 'hooks':
        if isinstance(value, str):
            is_valid_hook = valid_hook_type(value)
        elif isinstance(value, collections.abc.Sequence):
            is_valid_hook = all(valid_hook_type(hook) for hook in value)

        if not is_valid_hook:
            msg = (_('Invalid hook type "%(value)s" for resource '
                     'breakpoint, acceptable hook types are: %(types)s') %
                   {'value': value, 'types': HOOK_TYPES})
            raise exception.InvalidBreakPointHook(message=msg)

    return is_valid_hook


def is_valid_restricted_action(key, value):
    valid_action = False
    if key == 'restricted_actions':
        if isinstance(value, str):
            valid_action = valid_restricted_actions(value)
        elif isinstance(value, collections.abc.Sequence):
            valid_action = all(valid_restricted_actions(
                action) for action in value)

        if not valid_action:
            msg = (_('Invalid restricted_action type "%(value)s" for '
                     'resource, acceptable restricted_action '
                     'types are: %(types)s') %
                   {'value': value, 'types': RESTRICTED_ACTIONS})
            raise exception.InvalidRestrictedAction(message=msg)

    return valid_action


class ResourceInfo(object):
    """Base mapping of resource type to implementation."""

    def __new__(cls, registry, path, value):
        """Create a new ResourceInfo of the appropriate class."""

        if cls is not ResourceInfo:
            # Call is already for a subclass, so pass it through
            return super(ResourceInfo, cls).__new__(cls)

        name = path[-1]
        if name.endswith(('.yaml', '.template')):
            # a template url for the resource "Type"
            klass = TemplateResourceInfo
        elif not isinstance(value, str):
            klass = ClassResourceInfo
        elif value.endswith(('.yaml', '.template')):
            # a registered template
            klass = TemplateResourceInfo
        elif name.endswith('*'):
            klass = GlobResourceInfo
        else:
            klass = MapResourceInfo

        return super(ResourceInfo, cls).__new__(klass)

    __slots__ = ('_registry', 'path', 'name', 'value', 'user_resource')

    def __init__(self, registry, path, value):
        self._registry = weakref.ref(registry)
        self.path = path
        self.name = path[-1]
        self.value = value
        self.user_resource = True

    @property
    def registry(self):
        return self._registry()

    def __eq__(self, other):
        if other is None:
            return False
        return (self.path == other.path and
                self.value == other.value and
                self.user_resource == other.user_resource)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if self.user_resource != other.user_resource:
            # user resource must be sorted above system ones.
            return self.user_resource > other.user_resource
        if len(self.path) != len(other.path):
            # more specific (longer) path must be sorted above system ones.
            return len(self.path) > len(other.path)
        return self.path < other.path

    def __gt__(self, other):
        return other.__lt__(self)

    def get_resource_info(self, resource_type=None, resource_name=None):
        return self

    def matches(self, resource_type):
        return False

    def get_class(self):
        raise NotImplementedError

    def get_class_to_instantiate(self):
        return self.get_class()

    def __str__(self):
        return '[%s](User:%s) %s -> %s' % (self.description,
                                           self.user_resource,
                                           self.name, str(self.value))


class ClassResourceInfo(ResourceInfo):
    """Store the mapping of resource name to python class implementation."""
    description = 'Plugin'

    __slots__ = tuple()

    def get_class(self, files=None):
        return self.value


class TemplateResourceInfo(ResourceInfo):
    """Store the info needed to start a TemplateResource."""
    description = 'Template'

    __slots__ = ('template_name',)

    def __init__(self, registry, path, value):
        super(TemplateResourceInfo, self).__init__(registry, path, value)
        if self.name.endswith(('.yaml', '.template')):
            self.template_name = self.name
        else:
            self.template_name = value
        self.value = self.template_name

    def get_class(self, files=None):
        from heat.engine.resources import template_resource
        if files and self.template_name in files:
            data = files[self.template_name]
        else:
            if self.user_resource:
                allowed_schemes = template_resource.REMOTE_SCHEMES
            else:
                allowed_schemes = template_resource.LOCAL_SCHEMES
            data = template_resource.TemplateResource.get_template_file(
                self.template_name,
                allowed_schemes)
        param_defaults = self.registry.param_defaults
        return template_resource.generate_class_from_template(str(self.name),
                                                              data,
                                                              param_defaults)

    def get_class_to_instantiate(self):
        from heat.engine.resources import template_resource
        return template_resource.TemplateResource


class MapResourceInfo(ResourceInfo):
    """Store the mapping of one resource type to another.

    like: OS::Networking::FloatingIp -> OS::Neutron::FloatingIp
    """
    description = 'Mapping'

    __slots__ = tuple()

    def get_class(self, files=None):
        return None

    def get_resource_info(self, resource_type=None, resource_name=None):
        return self.registry.get_resource_info(self.value, resource_name)


class GlobResourceInfo(MapResourceInfo):
    """Store the mapping (with wild cards) of one resource type to another.

    like: OS::Networking::* -> OS::Neutron::*

    Also supports many-to-one mapping (mostly useful together with special
    "OS::Heat::None" resource)

    like: OS::* -> OS::Heat::None
    """
    description = 'Wildcard Mapping'

    __slots__ = tuple()

    def get_resource_info(self, resource_type=None, resource_name=None):
        # NOTE(pas-ha) we end up here only when self.name already
        # ends with * so truncate it
        orig_prefix = self.name[:-1]
        if self.value.endswith('*'):
            new_type = self.value[:-1] + resource_type[len(orig_prefix):]
        else:
            new_type = self.value

        return self.registry.get_resource_info(new_type, resource_name)

    def matches(self, resource_type):
        # prevent self-recursion in case of many-to-one mapping
        match = (resource_type != self.value and
                 resource_type.startswith(self.name[:-1]))
        return match


class ResourceRegistry(object):
    """By looking at the environment, find the resource implementation."""

    def __init__(self, global_registry, param_defaults):
        self._registry = {'resources': {}}
        self.global_registry = global_registry
        self.param_defaults = param_defaults

    def load(self, json_snippet):
        self._load_registry([], json_snippet)

    def register_class(self, resource_type, resource_class, path=None):
        if path is None:
            path = [resource_type]
        ri = ResourceInfo(self, path, resource_class)
        self._register_info(path, ri)

    def _load_registry(self, path, registry):
        for k, v in iter(registry.items()):
            if v is None:
                self._register_info(path + [k], None)
            elif is_hook_definition(k, v) or is_valid_restricted_action(k, v):
                self._register_item(path + [k], v)
            elif isinstance(v, dict):
                self._load_registry(path + [k], v)
            else:
                self._register_info(path + [k],
                                    ResourceInfo(self, path + [k], v))

    def _register_item(self, path, item):
        name = path[-1]
        registry = self._registry
        for key in path[:-1]:
            if key not in registry:
                registry[key] = {}
            registry = registry[key]
        registry[name] = item

    def _register_info(self, path, info):
        """Place the new info in the correct location in the registry.

        :param path: a list of keys ['resources', 'my_srv', 'OS::Nova::Server']
        """
        descriptive_path = '/'.join(path)
        name = path[-1]
        # create the structure if needed
        registry = self._registry
        for key in path[:-1]:
            if key not in registry:
                registry[key] = {}
            registry = registry[key]

        if info is None:
            if name.endswith('*'):
                # delete all matching entries.
                for res_name, reg_info in list(registry.items()):
                    if (isinstance(reg_info, ResourceInfo) and
                            res_name.startswith(name[:-1])):
                        LOG.warning('Removing %(item)s from %(path)s', {
                            'item': res_name,
                            'path': descriptive_path})
                        del registry[res_name]
            else:
                # delete this entry.
                LOG.warning('Removing %(item)s from %(path)s', {
                    'item': name,
                    'path': descriptive_path})
                registry.pop(name, None)
            return

        if name in registry and isinstance(registry[name], ResourceInfo):
            if registry[name] == info:
                return
            details = {
                'path': descriptive_path,
                'was': str(registry[name].value),
                'now': str(info.value)}
            LOG.warning('Changing %(path)s from %(was)s to %(now)s', details)

        if isinstance(info, ClassResourceInfo):
            if info.value.support_status.status != support.SUPPORTED:
                if info.value.support_status.message is not None:
                    details = {
                        'name': info.name,
                        'status': str(info.value.support_status.status),
                        'message': str(info.value.support_status.message)
                        }
                    LOG.warning('%(name)s is %(status)s. %(message)s',
                                details)

        info.user_resource = (self.global_registry is not None)
        registry[name] = info

    def log_resource_info(self, show_all=False, prefix=None):
        registry = self._registry
        prefix = '%s ' % prefix if prefix is not None else ''
        for name in registry:
            if name == 'resources':
                continue
            if show_all or isinstance(registry[name], TemplateResourceInfo):
                msg = ('%(p)sRegistered: %(t)s' %
                       {'p': prefix,
                        't': str(registry[name])})
                LOG.info(msg)

    def remove_item(self, info):
        if not isinstance(info, TemplateResourceInfo):
            return

        registry = self._registry
        for key in info.path[:-1]:
            registry = registry[key]
        if info.path[-1] in registry:
            registry.pop(info.path[-1])

    def get_rsrc_restricted_actions(self, resource_name):
        """Returns a set of restricted actions.

        For a given resource we get the set of restricted actions.

        Actions are set in this format via `resources`::

            {
                "restricted_actions": [update, replace]
            }

        A restricted_actions value is either `update`, `replace` or a list
        of those values. Resources support wildcard matching. The asterisk
        sign matches everything.
        """
        ress = self._registry['resources']
        restricted_actions = set()
        for name_pattern, resource in ress.items():
            if fnmatch.fnmatchcase(resource_name, name_pattern):
                if 'restricted_actions' in resource:
                    actions = resource['restricted_actions']
                    if isinstance(actions, str):
                        restricted_actions.add(actions)
                    elif isinstance(actions, collections.abc.Sequence):
                        restricted_actions |= set(actions)
        return restricted_actions

    def matches_hook(self, resource_name, hook):
        """Return whether a resource have a hook set in the environment.

        For a given resource and a hook type, we check to see if the passed
        group of resources has the right hook associated with the name.

        Hooks are set in this format via `resources`::

            {
                "res_name": {
                    "hooks": [pre-create, pre-update]
                },
                "*_suffix": {
                    "hooks": pre-create
                },
                "prefix_*": {
                    "hooks": pre-update
                }
            }

        A hook value is either `pre-create`, `pre-update` or a list of those
        values. Resources support wildcard matching. The asterisk sign matches
        everything.
        """
        ress = self._registry['resources']
        for name_pattern, resource in ress.items():
            if fnmatch.fnmatchcase(resource_name, name_pattern):
                if 'hooks' in resource:
                    hooks = resource['hooks']
                    if isinstance(hooks, str):
                        if hook == hooks:
                            return True
                    elif isinstance(hooks, collections.abc.Sequence):
                        if hook in hooks:
                            return True
        return False

    def remove_resources_except(self, resource_name):
        ress = self._registry['resources']
        new_resources = {}
        for name, res in ress.items():
            if fnmatch.fnmatchcase(resource_name, name):
                new_resources.update(res)
        if resource_name in ress:
            new_resources.update(ress[resource_name])
        self._registry['resources'] = new_resources

    def iterable_by(self, resource_type, resource_name=None):
        is_templ_type = resource_type.endswith(('.yaml', '.template'))
        if self.global_registry is not None and is_templ_type:
            # we only support dynamic resource types in user environments
            # not the global environment.
            # resource with a Type == a template
            # we dynamically create an entry as it has not been registered.
            if resource_type not in self._registry:
                res = ResourceInfo(self, [resource_type], None)
                self._register_info([resource_type], res)
            yield self._registry[resource_type]

        # handle a specific resource mapping.
        if resource_name:
            impl = self._registry['resources'].get(resource_name)
            if impl and resource_type in impl:
                yield impl[resource_type]

        # handle: "OS::Nova::Server" -> "Rackspace::Cloud::Server"
        impl = self._registry.get(resource_type)
        if impl:
            yield impl

        # handle: "OS::*" -> "Dreamhost::*"
        def is_a_glob(resource_type):
            return resource_type.endswith('*')
        globs = filter(is_a_glob, iter(self._registry))
        for pattern in globs:
            if self._registry[pattern].matches(resource_type):
                yield self._registry[pattern]

    def get_resource_info(self, resource_type, resource_name=None,
                          registry_type=None, ignore=None):
        """Find possible matches to the resource type and name.

        Chain the results from the global and user registry to find
        a match.
        """
        # use cases
        # 1) get the impl.
        #    - filter_by(res_type=X), sort_by(res_name=W, is_user=True)
        # 2) in TemplateResource we need to get both the
        #    TemplateClass and the ResourceClass
        #    - filter_by(res_type=X, impl_type=TemplateResourceInfo),
        #      sort_by(res_name=W, is_user=True)
        #    - filter_by(res_type=X, impl_type=ClassResourceInfo),
        #      sort_by(res_name=W, is_user=True)
        # 3) get_types() from the api
        #    - filter_by(is_user=False)
        # 4) as_dict() to write to the db
        #    - filter_by(is_user=True)

        if self.global_registry is not None:
            giter = self.global_registry.iterable_by(resource_type,
                                                     resource_name)
        else:
            giter = []

        matches = itertools.chain(self.iterable_by(resource_type,
                                                   resource_name),
                                  giter)

        for info in sorted(matches):
            try:
                match = info.get_resource_info(resource_type,
                                               resource_name)
            except exception.EntityNotFound:
                continue

            if registry_type is None or isinstance(match, registry_type):
                if ignore is not None and match == ignore:
                    continue
                # NOTE(prazumovsky): if resource_type defined in outer env
                # there is a risk to lose it due to h-eng restarting, so
                # store it to local env (exclude ClassResourceInfo because it
                # loads from resources; TemplateResourceInfo handles by
                # template_resource module).
                if (match and not match.user_resource and
                    not isinstance(info, (TemplateResourceInfo,
                                          ClassResourceInfo))):
                    self._register_info([resource_type], info)
                return match

        raise exception.EntityNotFound(entity='Resource Type',
                                       name=resource_type)

    def get_class(self, resource_type, resource_name=None, files=None):
        info = self.get_resource_info(resource_type,
                                      resource_name=resource_name)
        return info.get_class(files=files)

    def get_class_to_instantiate(self, resource_type, resource_name=None):
        if resource_type == "":
            msg = _('Resource "%s" has no type') % resource_name
            raise exception.StackValidationFailed(message=msg)
        elif resource_type is None:
            msg = _('Non-empty resource type is required '
                    'for resource "%s"') % resource_name
            raise exception.StackValidationFailed(message=msg)
        elif not isinstance(resource_type, str):
            msg = _('Resource "%s" type is not a string') % resource_name
            raise exception.StackValidationFailed(message=msg)

        try:
            info = self.get_resource_info(resource_type,
                                          resource_name=resource_name)
        except exception.EntityNotFound as exc:
            raise exception.StackValidationFailed(message=str(exc))

        return info.get_class_to_instantiate()

    def as_dict(self):
        """Return user resources in a dict format."""
        def _as_dict(level):
            tmp = {}
            for k, v in iter(level.items()):
                if isinstance(v, dict):
                    tmp[k] = _as_dict(v)
                elif is_hook_definition(
                        k, v) or is_valid_restricted_action(k, v):
                    tmp[k] = v
                elif v.user_resource:
                    tmp[k] = v.value
            return tmp

        return _as_dict(self._registry)

    def get_types(self,
                  cnxt=None,
                  support_status=None,
                  type_name=None,
                  version=None,
                  with_description=False):
        """Return a list of valid resource types."""

        # validate the support status
        if support_status is not None and not support.is_valid_status(
                support_status):
            msg = (_('Invalid support status and should be one of %s') %
                   str(support.SUPPORT_STATUSES))
            raise exception.Invalid(reason=msg)

        enforcer = policy.ResourceEnforcer()
        if type_name is not None:
            try:
                name_exp = re.compile(type_name)
            except Exception:
                return []
        else:
            name_exp = None

        def matches(name, info):
            # Only return actual plugins or template resources, not aliases
            if not isinstance(info, (ClassResourceInfo, TemplateResourceInfo)):
                return False

            # If filtering by name, check for match
            if name_exp is not None and not name_exp.match(name):
                return False

            rsrc_cls = info.get_class_to_instantiate()

            # Never match hidden resource types
            if rsrc_cls.support_status.status == support.HIDDEN:
                return False

            # If filtering by version, check for match
            if (version is not None and
                    rsrc_cls.support_status.version != version):
                return False

            # If filtering by support status, check for match
            if (support_status is not None and
                    rsrc_cls.support_status.status != support_status):
                return False

            if cnxt is not None:
                # Check for resource policy
                try:
                    enforcer.enforce(cnxt, name, is_registered_policy=True)
                except enforcer.exc:
                    return False

                # Check for service availability
                try:
                    avail, err = rsrc_cls.is_service_available(cnxt)
                except Exception:
                    avail = False
                if not avail:
                    return False

            return True

        import heat.engine.resource

        def resource_description(name, info):
            if not with_description:
                return name
            rsrc_cls = info.get_class()
            if rsrc_cls is None:
                rsrc_cls = heat.engine.resource.Resource
            return {
                'resource_type': name,
                'description': rsrc_cls.getdoc(),
            }

        return [resource_description(name, info)
                for name, info in self._registry.items()
                if matches(name, info)]


class Environment(object):

    def __init__(self, env=None, user_env=True):
        """Create an Environment from an input dict.

        The dict may be in one of two formats:
          1) old-school flat parameters; or
          2) newer {resource_registry: bla, parameters: foo}

        :param env: the json environment
        :param user_env: boolean, if False then we manage python resources too.
        """
        if env is None:
            env = {}
        if user_env:
            from heat.engine import resources
            global_env = resources.global_env()
            global_registry = global_env.registry
            event_sink_classes = global_env.event_sink_classes
        else:
            global_registry = None
            event_sink_classes = {}

        self.param_defaults = env.get(env_fmt.PARAMETER_DEFAULTS, {})

        self.registry = ResourceRegistry(global_registry, self.param_defaults)
        self.registry.load(env.get(env_fmt.RESOURCE_REGISTRY, {}))

        self.encrypted_param_names = env.get(env_fmt.ENCRYPTED_PARAM_NAMES, [])

        if env_fmt.PARAMETERS in env:
            self.params = env[env_fmt.PARAMETERS]
        else:
            self.params = dict((k, v) for (k, v) in env.items()
                               if k not in (env_fmt.PARAMETER_DEFAULTS,
                                            env_fmt.ENCRYPTED_PARAM_NAMES,
                                            env_fmt.EVENT_SINKS,
                                            env_fmt.RESOURCE_REGISTRY))
        self.event_sink_classes = event_sink_classes
        self._event_sinks = []
        self._built_event_sinks = []
        self._update_event_sinks(env.get(env_fmt.EVENT_SINKS, []))
        self.constraints = {}
        self.stack_lifecycle_plugins = []

    def load(self, env_snippet):
        self.registry.load(env_snippet.get(env_fmt.RESOURCE_REGISTRY, {}))
        self.params.update(env_snippet.get(env_fmt.PARAMETERS, {}))
        self.param_defaults.update(
            env_snippet.get(env_fmt.PARAMETER_DEFAULTS, {}))
        self._update_event_sinks(env_snippet.get(env_fmt.EVENT_SINKS, []))

    def env_as_dict(self):
        """Get the entire environment as a dict."""
        user_env = self.user_env_as_dict()
        user_env.update(
            # Any data here is to be stored in the DB but not reflected
            # as part of the user environment (e.g to pass to nested stacks
            # or made visible to the user via API calls etc
            {env_fmt.ENCRYPTED_PARAM_NAMES: self.encrypted_param_names})
        return user_env

    def user_env_as_dict(self):
        """Get the environment as a dict, only user-allowed keys."""
        return {env_fmt.RESOURCE_REGISTRY: self.registry.as_dict(),
                env_fmt.PARAMETERS: self.params,
                env_fmt.PARAMETER_DEFAULTS: self.param_defaults,
                env_fmt.EVENT_SINKS: self._event_sinks}

    def register_class(self, resource_type, resource_class, path=None):
        self.registry.register_class(resource_type, resource_class, path=path)

    def register_constraint(self, constraint_name, constraint):
        self.constraints[constraint_name] = constraint

    def register_stack_lifecycle_plugin(self, stack_lifecycle_name,
                                        stack_lifecycle_class):
        self.stack_lifecycle_plugins.append((stack_lifecycle_name,
                                             stack_lifecycle_class))

    def register_event_sink(self, event_sink_name, event_sink_class):
        self.event_sink_classes[event_sink_name] = event_sink_class

    def get_class(self, resource_type, resource_name=None, files=None):
        return self.registry.get_class(resource_type, resource_name,
                                       files=files)

    def get_class_to_instantiate(self, resource_type, resource_name=None):
        return self.registry.get_class_to_instantiate(resource_type,
                                                      resource_name)

    def get_types(self,
                  cnxt=None,
                  support_status=None,
                  type_name=None,
                  version=None,
                  with_description=False):
        return self.registry.get_types(cnxt,
                                       support_status=support_status,
                                       type_name=type_name,
                                       version=version,
                                       with_description=with_description)

    def get_resource_info(self, resource_type, resource_name=None,
                          registry_type=None, ignore=None):
        return self.registry.get_resource_info(resource_type, resource_name,
                                               registry_type, ignore=ignore)

    def get_constraint(self, name):
        return self.constraints.get(name)

    def get_stack_lifecycle_plugins(self):
        return self.stack_lifecycle_plugins

    def _update_event_sinks(self, sinks):
        self._event_sinks.extend(sinks)
        for sink in sinks:
            sink = sink.copy()
            sink_class = sink.pop('type')
            sink_class = self.event_sink_classes[sink_class]
            self._built_event_sinks.append(sink_class(**sink))

    def get_event_sinks(self):
        return self._built_event_sinks


def get_child_environment(parent_env, child_params, item_to_remove=None,
                          child_resource_name=None):
    """Build a child environment using the parent environment and params.

    This is built from the child_params and the parent env so some
    resources can use user-provided parameters as if they come from an
    environment.

    1. resource_registry must be merged (child env should be loaded after the
       parent env to take precedence).
    2. child parameters must overwrite the parent's as they won't be relevant
       in the child template.

    If `child_resource_name` is provided, resources in the registry will be
    replaced with the contents of the matching child resource plus anything
    that passes a wildcard match.
    """
    def is_flat_params(env_or_param):
        if env_or_param is None:
            return False
        for sect in env_fmt.SECTIONS:
            if sect in env_or_param:
                return False
        return True

    child_env = parent_env.user_env_as_dict()
    child_env[env_fmt.PARAMETERS] = {}
    flat_params = is_flat_params(child_params)
    new_env = Environment()
    if flat_params and child_params is not None:
        child_env[env_fmt.PARAMETERS] = child_params

    new_env.load(child_env)
    if not flat_params and child_params is not None:
        new_env.load(child_params)

    if item_to_remove is not None:
        new_env.registry.remove_item(item_to_remove)

    if child_resource_name:
        new_env.registry.remove_resources_except(child_resource_name)
    return new_env


def read_global_environment(env, env_dir=None):
    if env_dir is None:
        cfg.CONF.import_opt('environment_dir', 'heat.common.config')
        env_dir = cfg.CONF.environment_dir

    try:
        env_files = glob.glob(os.path.join(env_dir, '*'))
    except OSError:
        LOG.exception('Failed to read %s', env_dir)
        return

    for file_path in env_files:
        try:
            with open(file_path) as env_fd:
                LOG.info('Loading %s', file_path)
                env_body = env_fmt.parse(env_fd.read())
                env_fmt.default_for_missing(env_body)
                env.load(env_body)
        except ValueError:
            LOG.exception('Failed to parse %s', file_path)
        except IOError:
            LOG.exception('Failed to read %s', file_path)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json

from heat.common import exception
from heat.common import short_id
from heat.common import template_format
from heat.engine import environment
from heat.engine import stack as parser
from heat.engine import template
from heat.tests import common
from heat.tests import utils


class TestValue(common.HeatTestCase):

    simple_template = '''
heat_template_version: '2016-10-14'
parameters:
  param1:
    type: <the type>
resources:
  my_value:
    type: OS::Heat::Value
    properties:
      value: {get_param: param1}
  my_value2:
    type: OS::Heat::Value
    properties:
      value: {get_attr: [my_value, value]}
outputs:
  myout:
    value: {get_attr: [my_value2, value]}
'''

    def get_strict_and_loose_templates(self, param_type):
        template_loose = template_format.parse(self.simple_template)
        template_loose['parameters']['param1']['type'] = param_type
        template_strict = copy.deepcopy(template_loose)
        template_strict['resources']['my_value']['properties']['type'] \
            = param_type
        template_strict['resources']['my_value2']['properties']['type'] \
            = param_type
        return (template_strict, template_loose)

    def parse_stack(self, templ_obj):
        stack_name = 'test_value_stack_%s' % short_id.generate_id()
        stack = parser.Stack(utils.dummy_context(), stack_name, templ_obj)
        stack.validate()
        stack.store()
        return stack

    def create_stack(self, templ, env=None):
        if isinstance(templ, str):
            return self.create_stack(template_format.parse(templ), env=env)
        if isinstance(templ, dict):
            tmpl_obj = template.Template(templ, env=env)
            return self.create_stack(tmpl_obj)
        assert isinstance(templ, template.Template)
        stack = self.parse_stack(templ)
        self.assertIsNone(stack.create())
        return stack


class TestValueSimple(TestValue):

    scenarios = [
        ('boolean', dict(
            param1=True, param_type="boolean")),
        ('list', dict(
            param1=['a', 'b', 'Z'], param_type="comma_delimited_list")),
        ('map', dict(
            param1={'a': 'Z', 'B': 'y'}, param_type="json")),
        ('number-int', dict(
            param1=-11, param_type="number")),
        ('number-float', dict(
            param1=100.999, param_type="number")),
        ('string', dict(
            param1='Perchance to dream', param_type="string")),
    ]

    def test_value(self):
        ts, tl = self.get_strict_and_loose_templates(self.param_type)
        env = environment.Environment({
            'parameters': {'param1': self.param1}})

        for templ_dict in [ts, tl]:
            stack = self.create_stack(templ_dict, env)
            self.assertEqual(self.param1, stack['my_value'].FnGetAtt('value'))
            self.assertEqual(self.param1, stack['my_value2'].FnGetAtt('value'))
            stack._update_all_resource_data(False, True)
            self.assertEqual(self.param1, stack.outputs['myout'].get_value())


class TestValueLessSimple(TestValue):

    template_bad = '''
heat_template_version: '2016-10-14'
parameters:
  param1:
    type: json
resources:
  my_value:
    type: OS::Heat::Value
    properties:
      value: {get_param: param1}
      type: number
'''

    template_map = '''
heat_template_version: '2016-10-14'
parameters:
  param1:
    type: json
  param2:
    type: json
resources:
  my_value:
    type: OS::Heat::Value
    properties:
      value: {get_param: param1}
      type: json
  my_value2:
    type: OS::Heat::Value
    properties:
      value: {map_merge: [{get_attr: [my_value, value]}, {get_param: param2}]}
      type: json
'''

    template_yaql = '''
heat_template_version: '2016-10-14'
parameters:
  param1:
    type: number
  param2:
    type: comma_delimited_list
resources:
  my_value:
    type: OS::Heat::Value
    properties:
      value: {get_param: param1}
      type: number
  my_value2:
    type: OS::Heat::Value
    properties:
      value:
        yaql:
          expression: $.data.param2.select(int($)).min()
          data:
            param2: {get_param: param2}
      type: number
  my_value3:
    type: OS::Heat::Value
    properties:
      value:
        yaql:
          expression: min($.data.v1,$.data.v2)
          data:
            v1: {get_attr: [my_value, value]}
            v2: {get_attr: [my_value2, value]}
'''

    def test_validation_fail(self):
        param1 = {"one": "croissant"}
        env = environment.Environment({
            'parameters': {'param1': json.dumps(param1)}})
        self.assertRaises(exception.StackValidationFailed,
                          self.create_stack, self.template_bad, env)

    def test_map(self):
        param1 = {"one": "skipper", "two": "antennae"}
        param2 = {"one": "monarch", "three": "sky"}
        env = environment.Environment({
            'parameters': {'param1': json.dumps(param1),
                           'param2': json.dumps(param2)}})
        stack = self.create_stack(self.template_map, env)
        my_value = stack['my_value']
        self.assertEqual(param1, my_value.FnGetAtt('value'))
        my_value2 = stack['my_value2']
        self.assertEqual({"one": "monarch",
                          "two": "antennae",
                          "three": "sky"}, my_value2.FnGetAtt('value'))

    def test_yaql(self):
        param1 = -800
        param2 = [-8, 0, 4, -11, 2]
        env = environment.Environment({
            'parameters': {'param1': param1, 'param2': param2}})
        stack = self.create_stack(self.template_yaql, env)
        my_value = stack['my_value']
        self.assertEqual(param1, my_value.FnGetAtt('value'))
        my_value2 = stack['my_value2']
        self.assertEqual(min(param2), my_value2.FnGetAtt('value'))
        my_value3 = stack['my_value3']
        self.assertEqual(param1, my_value3.FnGetAtt('value'))


class TestValueUpdate(TestValue):

    scenarios = [
        ('boolean-to-number', dict(
            param1=True, param_type1="boolean",
            param2=-100.999, param_type2="number")),
        ('number-to-string', dict(
            param1=-77, param_type1="number",
            param2='mellors', param_type2="string")),
        ('string-to-map', dict(
            param1='mellors', param_type1="string",
            param2={'3': 'turbo'}, param_type2="json")),
        ('map-to-boolean', dict(
            param1={'hey': 'there'}, param_type1="json",
            param2=False, param_type2="boolean")),
        ('list-to-boolean', dict(
            param1=['hey', '!'], param_type1="comma_delimited_list",
            param2=True, param_type2="boolean")),
    ]

    def test_value_update(self):
        ts1, tl1 = self.get_strict_and_loose_templates(self.param_type1)
        ts2, tl2 = self.get_strict_and_loose_templates(self.param_type2)

        env1 = environment.Environment({
            'parameters': {'param1': self.param1}})
        env2 = environment.Environment({
            'parameters': {'param1': self.param2}})

        updates = [(ts1, ts2), (ts1, tl2), (tl1, ts2), (tl1, tl2)]
        updates_other_way = [(b, a) for a, b in updates]
        updates.extend(updates_other_way)
        for t_initial, t_updated in updates:
            if t_initial == ts1 or t_initial == tl1:
                p1, p2, e1, e2 = self.param1, self.param2, env1, env2
            else:
                # starting with param2, updating to param1
                p2, p1, e2, e1 = self.param1, self.param2, env1, env2
            stack = self.create_stack(copy.deepcopy(t_initial), env=e1)
            self.assertEqual(p1, stack['my_value2'].FnGetAtt('value'))
            res1_id = stack['my_value'].id
            res2_id = stack['my_value2'].id
            res2_uuid = stack['my_value2'].uuid

            updated_stack = parser.Stack(
                stack.context, 'updated_stack',
                template.Template(copy.deepcopy(t_updated), env=e2))
            updated_stack.validate()
            stack.update(updated_stack)
            self.assertEqual(p2, stack['my_value2'].FnGetAtt('value'))
            # Make sure resources not replaced after update
            self.assertEqual(res1_id, stack['my_value'].id)
            self.assertEqual(res2_id, stack['my_value2'].id)
            self.assertEqual(res2_uuid, stack['my_value2'].uuid)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Heat API Server.

This implements an approximation of the Amazon CloudFormation API and
translates it into a native representation.  It then calls the heat-engine via
AMQP RPC to implement them.
"""

# flake8: noqa: E402

import eventlet
eventlet.monkey_patch(os=False)

import sys

from oslo_config import cfg
import oslo_i18n as i18n
from oslo_log import log as logging
from oslo_reports import guru_meditation_report as gmr
from oslo_reports import opts as gmr_opts
from oslo_service import systemd

from heat.common import config
from heat.common import messaging
from heat.common import profiler
from heat.common import wsgi
from heat import version


i18n.enable_lazy()

CONF = cfg.CONF


def launch_cfn_api(setup_logging=True):
    if setup_logging:
        logging.register_options(CONF)
    CONF(project='heat',
         prog='heat-api-cfn',
         version=version.version_info.version_string())
    if setup_logging:
        logging.setup(CONF, CONF.prog)
        logging.set_defaults()
    LOG = logging.getLogger(CONF.prog)
    config.set_config_defaults()
    messaging.setup()

    app = config.load_paste_app()

    port = CONF.heat_api_cfn.bind_port
    host = CONF.heat_api_cfn.bind_host
    LOG.info('Starting Heat API on %(host)s:%(port)s',
             {'host': host, 'port': port})
    profiler.setup(CONF.prog, host)
    gmr_opts.set_defaults(CONF)
    gmr.TextGuruMeditation.setup_autorun(version, conf=CONF)
    server = wsgi.Server(CONF.prog, CONF.heat_api_cfn)
    server.start(app, default_port=port)
    return server


def main():
    try:
        server = launch_cfn_api()
        systemd.notify_once()
        server.wait()
    except RuntimeError as e:
        msg = str(e)
        sys.exit("ERROR: %s" % msg)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils as json

from heat.common import exception
from heat.common import identifier
from heat.engine import parameters
from heat.engine import template
from heat.tests import common


def new_parameter(name, schema, value=None, validate_value=True):
    tmpl = template.Template({'HeatTemplateFormatVersion': '2012-12-12',
                              'Parameters': {name: schema}})
    schema = tmpl.param_schemata()[name]
    param = parameters.Parameter(name, schema, value)
    param.validate(validate_value)
    return param


class ParameterTestCommon(common.HeatTestCase):
    scenarios = [
        ('type_string', dict(p_type='String',
                             inst=parameters.StringParam,
                             value='test',
                             expected='test',
                             allowed_value=['foo'],
                             zero='',
                             default='default')),
        ('type_number', dict(p_type='Number',
                             inst=parameters.NumberParam,
                             value=10,
                             expected='10',
                             allowed_value=[42],
                             zero=0,
                             default=13)),
        ('type_list', dict(p_type='CommaDelimitedList',
                           inst=parameters.CommaDelimitedListParam,
                           value=['a', 'b', 'c'],
                           expected='a,b,c',
                           allowed_value=['foo'],
                           zero=[],
                           default=['d', 'e', 'f'])),
        ('type_json', dict(p_type='Json',
                           inst=parameters.JsonParam,
                           value={'a': '1'},
                           expected='{"a": "1"}',
                           allowed_value=[{'foo': 'bar'}],
                           zero={},
                           default={'d': '1'})),
        ('type_int_json', dict(p_type='Json',
                               inst=parameters.JsonParam,
                               value={'a': 1},
                               expected='{"a": 1}',
                               allowed_value=[{'foo': 'bar'}],
                               zero={},
                               default={'d': 1})),
        ('type_boolean', dict(p_type='Boolean',
                              inst=parameters.BooleanParam,
                              value=True,
                              expected='True',
                              allowed_value=[False],
                              zero=False,
                              default=True)),
        ('type_int_string', dict(p_type='String',
                                 inst=parameters.StringParam,
                                 value='111',
                                 expected='111',
                                 allowed_value=['111'],
                                 zero='',
                                 default='0')),
        ('type_string_json', dict(p_type='Json',
                                  inst=parameters.JsonParam,
                                  value={'1': 1},
                                  expected='{"1": 1}',
                                  allowed_value=[{'2': '2'}],
                                  zero={},
                                  default={'3': 3}))
    ]

    def test_new_param(self):
        p = new_parameter('p', {'Type': self.p_type}, validate_value=False)
        self.assertIsInstance(p, self.inst)

    def test_param_to_str(self):
        p = new_parameter('p', {'Type': self.p_type}, self.value)
        if self.p_type == 'Json':
            self.assertEqual(json.loads(self.expected), json.loads(str(p)))
        else:
            self.assertEqual(self.expected, str(p))

    def test_default_no_override(self):
        p = new_parameter('defaulted', {'Type': self.p_type,
                                        'Default': self.default})
        self.assertTrue(p.has_default())
        self.assertEqual(self.default, p.default())
        self.assertEqual(self.default, p.value())

    def test_default_override(self):
        p = new_parameter('defaulted', {'Type': self.p_type,
                                        'Default': self.default},
                          self.value)
        self.assertTrue(p.has_default())
        self.assertEqual(self.default, p.default())
        self.assertEqual(self.value, p.value())

    def test_default_invalid(self):
        schema = {'Type': self.p_type,
                  'AllowedValues': self.allowed_value,
                  'ConstraintDescription': 'wibble',
                  'Default': self.default}
        if self.p_type == 'Json':
            err = self.assertRaises(exception.InvalidSchemaError,
                                    new_parameter, 'p', schema)
            self.assertIn('AllowedValues constraint invalid for Json',
                          str(err))
        else:
            err = self.assertRaises(exception.InvalidSchemaError,
                                    new_parameter, 'p', schema)
            self.assertIn('wibble', str(err))

    def test_description(self):
        description = 'Description of the parameter'
        p = new_parameter('p', {'Type': self.p_type,
                                'Description': description},
                          validate_value=False)
        self.assertEqual(description, p.description())

    def test_no_description(self):
        p = new_parameter('p', {'Type': self.p_type}, validate_value=False)
        self.assertEqual('', p.description())

    def test_no_echo_true(self):
        p = new_parameter('anechoic', {'Type': self.p_type,
                                       'NoEcho': 'true'},
                          self.value)
        self.assertTrue(p.hidden())
        self.assertEqual('******', str(p))

    def test_no_echo_true_caps(self):
        p = new_parameter('anechoic', {'Type': self.p_type,
                                       'NoEcho': 'TrUe'},
                          self.value)
        self.assertTrue(p.hidden())
        self.assertEqual('******', str(p))

    def test_no_echo_false(self):
        p = new_parameter('echoic', {'Type': self.p_type,
                                     'NoEcho': 'false'},
                          self.value)
        self.assertFalse(p.hidden())
        if self.p_type == 'Json':
            self.assertEqual(json.loads(self.expected), json.loads(str(p)))
        else:
            self.assertEqual(self.expected, str(p))

    def test_default_empty(self):
        p = new_parameter('defaulted', {'Type': self.p_type,
                                        'Default': self.zero})
        self.assertTrue(p.has_default())
        self.assertEqual(self.zero, p.default())
        self.assertEqual(self.zero, p.value())

    def test_default_no_empty_user_value_empty(self):
        p = new_parameter('defaulted', {'Type': self.p_type,
                                        'Default': self.default},
                          self.zero)
        self.assertTrue(p.has_default())
        self.assertEqual(self.default, p.default())
        self.assertEqual(self.zero, p.value())


class ParameterTestSpecific(common.HeatTestCase):
    def test_new_bad_type(self):
        self.assertRaises(exception.InvalidSchemaError, new_parameter,
                          'p', {'Type': 'List'}, validate_value=False)

    def test_string_len_good(self):
        schema = {'Type': 'String',
                  'MinLength': '3',
                  'MaxLength': '3'}
        p = new_parameter('p', schema, 'foo')
        self.assertEqual('foo', p.value())

    def test_string_underflow(self):
        schema = {'Type': 'String',
                  'ConstraintDescription': 'wibble',
                  'MinLength': '4'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, 'foo')
        self.assertIn('wibble', str(err))

    def test_string_overflow(self):
        schema = {'Type': 'String',
                  'ConstraintDescription': 'wibble',
                  'MaxLength': '2'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, 'foo')
        self.assertIn('wibble', str(err))

    def test_string_pattern_good(self):
        schema = {'Type': 'String',
                  'AllowedPattern': '[a-z]*'}
        p = new_parameter('p', schema, 'foo')
        self.assertEqual('foo', p.value())

    def test_string_pattern_bad_prefix(self):
        schema = {'Type': 'String',
                  'ConstraintDescription': 'wibble',
                  'AllowedPattern': '[a-z]*'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, '1foo')
        self.assertIn('wibble', str(err))

    def test_string_pattern_bad_suffix(self):
        schema = {'Type': 'String',
                  'ConstraintDescription': 'wibble',
                  'AllowedPattern': '[a-z]*'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, 'foo1')
        self.assertIn('wibble', str(err))

    def test_string_value_list_good(self):
        schema = {'Type': 'String',
                  'AllowedValues': ['foo', 'bar', 'baz']}
        p = new_parameter('p', schema, 'bar')
        self.assertEqual('bar', p.value())

    def test_string_value_unicode(self):
        schema = {'Type': 'String'}
        p = new_parameter('p', schema, 'test\u2665')
        self.assertEqual('test\u2665', p.value())

    def test_string_value_list_bad(self):
        schema = {'Type': 'String',
                  'ConstraintDescription': 'wibble',
                  'AllowedValues': ['foo', 'bar', 'baz']}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, 'blarg')
        self.assertIn('wibble', str(err))

    def test_number_int_good(self):
        schema = {'Type': 'Number',
                  'MinValue': '3',
                  'MaxValue': '3'}
        p = new_parameter('p', schema, '3')
        self.assertEqual(3, p.value())

    def test_number_float_good_string(self):
        schema = {'Type': 'Number',
                  'MinValue': '3.0',
                  'MaxValue': '4.0'}
        p = new_parameter('p', schema, '3.5')
        self.assertEqual(3.5, p.value())

    def test_number_float_good_number(self):
        schema = {'Type': 'Number',
                  'MinValue': '3.0',
                  'MaxValue': '4.0'}
        p = new_parameter('p', schema, 3.5)
        self.assertEqual(3.5, p.value())

    def test_number_low(self):
        schema = {'Type': 'Number',
                  'ConstraintDescription': 'wibble',
                  'MinValue': '4'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, '3')
        self.assertIn('wibble', str(err))

    def test_number_high(self):
        schema = {'Type': 'Number',
                  'ConstraintDescription': 'wibble',
                  'MaxValue': '2'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, '3')
        self.assertIn('wibble', str(err))

    def test_number_bad(self):
        schema = {'Type': 'Number'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, 'str')
        self.assertIn('float', str(err))

    def test_number_bad_type(self):
        schema = {'Type': 'Number'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, ['foo'])
        self.assertIn('int', str(err))

    def test_number_value_list_good(self):
        schema = {'Type': 'Number',
                  'AllowedValues': ['1', '3', '5']}
        p = new_parameter('p', schema, '5')
        self.assertEqual(5, p.value())

    def test_number_value_list_bad(self):
        schema = {'Type': 'Number',
                  'ConstraintDescription': 'wibble',
                  'AllowedValues': ['1', '3', '5']}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, '2')
        self.assertIn('wibble', str(err))

    def test_list_value_list_default_empty(self):
        schema = {'Type': 'CommaDelimitedList', 'Default': ''}
        p = new_parameter('p', schema)
        self.assertEqual([], p.value())

    def test_list_value_list_good(self):
        schema = {'Type': 'CommaDelimitedList',
                  'AllowedValues': ['foo', 'bar', 'baz']}
        p = new_parameter('p', schema, 'baz,foo,bar')
        self.assertEqual('baz,foo,bar'.split(','), p.value())
        schema['Default'] = []
        p = new_parameter('p', schema)
        self.assertEqual([], p.value())
        schema['Default'] = 'baz,foo,bar'
        p = new_parameter('p', schema)
        self.assertEqual('baz,foo,bar'.split(','), p.value())
        schema['AllowedValues'] = ['1', '3', '5']
        schema['Default'] = []
        p = new_parameter('p', schema, [1, 3, 5])
        self.assertEqual('1,3,5', str(p))
        schema['Default'] = [1, 3, 5]
        p = new_parameter('p', schema)
        self.assertEqual('1,3,5'.split(','), p.value())

    def test_list_value_list_bad(self):
        schema = {'Type': 'CommaDelimitedList',
                  'ConstraintDescription': 'wibble',
                  'AllowedValues': ['foo', 'bar', 'baz']}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema,
                                'foo,baz,blarg')
        self.assertIn('wibble', str(err))

    def test_list_validate_good(self):
        schema = {'Type': 'CommaDelimitedList'}
        val = ['foo', 'bar', 'baz']
        val_s = 'foo,bar,baz'
        p = new_parameter('p', schema, val_s, validate_value=False)
        p.validate()
        self.assertEqual(val, p.value())
        self.assertEqual(val, p.parsed)

    def test_list_validate_bad(self):
        schema = {'Type': 'CommaDelimitedList'}
        # just need something here that is growing to throw an AttributeError
        # when .split() is called
        val_s = 0
        p = new_parameter('p', schema, validate_value=False)
        p.user_value = val_s
        err = self.assertRaises(exception.StackValidationFailed,
                                p.validate)
        self.assertIn('Parameter \'p\' is invalid', str(err))

    def test_map_value(self):
        '''Happy path for value that's already a map.'''
        schema = {'Type': 'Json'}
        val = {"foo": "bar", "items": [1, 2, 3]}
        p = new_parameter('p', schema, val)
        self.assertEqual(val, p.value())
        self.assertEqual(val, p.parsed)

    def test_map_value_bad(self):
        '''Map value is not JSON parsable.'''
        schema = {'Type': 'Json',
                  'ConstraintDescription': 'wibble'}
        val = {"foo": "bar", "not_json": len}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, val)
        self.assertIn('Value must be valid JSON', str(err))

    def test_map_value_parse(self):
        '''Happy path for value that's a string.'''
        schema = {'Type': 'Json'}
        val = {"foo": "bar", "items": [1, 2, 3]}
        val_s = json.dumps(val)
        p = new_parameter('p', schema, val_s)
        self.assertEqual(val, p.value())
        self.assertEqual(val, p.parsed)

    def test_map_value_bad_parse(self):
        '''Test value error for unparsable string value.'''
        schema = {'Type': 'Json',
                  'ConstraintDescription': 'wibble'}
        val = "I am not a map"
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, val)
        self.assertIn('Value must be valid JSON', str(err))

    def test_map_underrun(self):
        '''Test map length under MIN_LEN.'''
        schema = {'Type': 'Json',
                  'MinLength': 3}
        val = {"foo": "bar", "items": [1, 2, 3]}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, val)
        self.assertIn('out of range', str(err))

    def test_map_overrun(self):
        '''Test map length over MAX_LEN.'''
        schema = {'Type': 'Json',
                  'MaxLength': 1}
        val = {"foo": "bar", "items": [1, 2, 3]}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'p', schema, val)
        self.assertIn('out of range', str(err))

    def test_json_list(self):
        schema = {'Type': 'Json'}
        val = ["fizz", "buzz"]
        p = new_parameter('p', schema, val)
        self.assertIsInstance(p.value(), list)
        self.assertIn("fizz", p.value())
        self.assertIn("buzz", p.value())

    def test_json_string_list(self):
        schema = {'Type': 'Json'}
        val = '["fizz", "buzz"]'
        p = new_parameter('p', schema, val)
        self.assertIsInstance(p.value(), list)
        self.assertIn("fizz", p.value())
        self.assertIn("buzz", p.value())

    def test_json_validate_good(self):
        schema = {'Type': 'Json'}
        val = {"foo": "bar", "items": [1, 2, 3]}
        val_s = json.dumps(val)
        p = new_parameter('p', schema, val_s, validate_value=False)
        p.validate()
        self.assertEqual(val, p.value())
        self.assertEqual(val, p.parsed)

    def test_json_validate_bad(self):
        schema = {'Type': 'Json'}
        val_s = '{"foo": "bar", "invalid": ]}'
        p = new_parameter('p', schema, validate_value=False)
        p.user_value = val_s
        err = self.assertRaises(exception.StackValidationFailed,
                                p.validate)
        self.assertIn('Parameter \'p\' is invalid', str(err))

    def test_bool_value_true(self):
        schema = {'Type': 'Boolean'}
        for val in ('1', 't', 'true', 'on', 'y', 'yes', True, 1):
            bo = new_parameter('bo', schema, val)
            self.assertTrue(bo.value())

    def test_bool_value_false(self):
        schema = {'Type': 'Boolean'}
        for val in ('0', 'f', 'false', 'off', 'n', 'no', False, 0):
            bo = new_parameter('bo', schema, val)
            self.assertFalse(bo.value())

    def test_bool_value_invalid(self):
        schema = {'Type': 'Boolean'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'bo', schema, 'foo')
        self.assertIn("Unrecognized value 'foo'", str(err))

    def test_missing_param_str(self):
        '''Test missing user parameter.'''
        self.assertRaises(exception.UserParameterMissing,
                          new_parameter, 'p',
                          {'Type': 'String'})

    def test_missing_param_list(self):
        '''Test missing user parameter.'''
        self.assertRaises(exception.UserParameterMissing,
                          new_parameter, 'p',
                          {'Type': 'CommaDelimitedList'})

    def test_missing_param_map(self):
        '''Test missing user parameter.'''
        self.assertRaises(exception.UserParameterMissing,
                          new_parameter, 'p',
                          {'Type': 'Json'})

    def test_param_name_in_error_message(self):
        schema = {'Type': 'String',
                  'AllowedPattern': '[a-z]*'}
        err = self.assertRaises(exception.StackValidationFailed,
                                new_parameter, 'testparam', schema, '234')
        expected = ("Parameter 'testparam' is invalid: "
                    '"234" does not match pattern "[a-z]*"')
        self.assertEqual(expected, str(err))


params_schema = json.loads('''{
  "Parameters" : {
    "User" : { "Type": "String" },
    "Defaulted" : {
      "Type": "String",
      "Default": "foobar"
    }
  }
}''')


class ParametersBase(common.HeatTestCase):
    def new_parameters(self, stack_name, tmpl, user_params=None,
                       stack_id=None, validate_value=True,
                       param_defaults=None):
        user_params = user_params or {}
        tmpl.update({'HeatTemplateFormatVersion': '2012-12-12'})
        tmpl = template.Template(tmpl)
        params = tmpl.parameters(
            identifier.HeatIdentifier('', stack_name, stack_id),
            user_params, param_defaults=param_defaults)
        params.validate(validate_value)
        return params


class ParametersTest(ParametersBase):

    def test_pseudo_params(self):
        stack_name = 'test_stack'
        params = self.new_parameters(stack_name, {"Parameters": {}})

        self.assertEqual('test_stack', params['AWS::StackName'])
        self.assertEqual(
            'arn:openstack:heat:::stacks/{0}/{1}'.format(stack_name, 'None'),
            params['AWS::StackId'])

        self.assertIn('AWS::Region', params)

    def test_pseudo_param_stackid(self):
        stack_name = 'test_stack'
        params = self.new_parameters(stack_name, {'Parameters': {}},
                                     stack_id='abc123')

        self.assertEqual(
            'arn:openstack:heat:::stacks/{0}/{1}'.format(stack_name, 'abc123'),
            params['AWS::StackId'])
        stack_identifier = identifier.HeatIdentifier('', '', 'def456')
        params.set_stack_id(stack_identifier)
        self.assertEqual(stack_identifier.arn(), params['AWS::StackId'])

    def test_schema_invariance(self):
        params1 = self.new_parameters('test', params_schema,
                                      {'User': 'foo',
                                       'Defaulted': 'wibble'})
        self.assertEqual('wibble', params1['Defaulted'])

        params2 = self.new_parameters('test', params_schema, {'User': 'foo'})
        self.assertEqual('foobar', params2['Defaulted'])

    def test_to_dict(self):
        template = {'Parameters': {'Foo': {'Type': 'String'},
                                   'Bar': {'Type': 'Number', 'Default': '42'}}}
        params = self.new_parameters('test_params', template, {'Foo': 'foo'})

        as_dict = dict(params)
        self.assertEqual('foo', as_dict['Foo'])
        self.assertEqual(42, as_dict['Bar'])
        self.assertEqual('test_params', as_dict['AWS::StackName'])
        self.assertIn('AWS::Region', as_dict)

    def test_map(self):
        template = {'Parameters': {'Foo': {'Type': 'String'},
                                   'Bar': {'Type': 'Number', 'Default': '42'}}}
        params = self.new_parameters('test_params', template, {'Foo': 'foo'})

        expected = {'Foo': False,
                    'Bar': True,
                    'AWS::Region': True,
                    'AWS::StackId': True,
                    'AWS::StackName': True}

        self.assertEqual(expected, params.map(lambda p: p.has_default()))

    def test_map_str(self):
        template = {'Parameters': {'Foo': {'Type': 'String'},
                                   'Bar': {'Type': 'Number'},
                                   'Uni': {'Type': 'String'}}}
        stack_name = 'test_params'
        params = self.new_parameters(stack_name, template,
                                     {'Foo': 'foo',
                                      'Bar': '42',
                                      'Uni': u'test\u2665'})

        expected = {'Foo': 'foo',
                    'Bar': '42',
                    'Uni': b'test\xe2\x99\xa5',
                    'AWS::Region': 'ap-southeast-1',
                    'AWS::StackId':
                    'arn:openstack:heat:::stacks/{0}/{1}'.format(
                        stack_name,
                        'None'),
                    'AWS::StackName': 'test_params'}

        mapped_params = params.map(str)
        mapped_params['Uni'] = mapped_params['Uni'].encode('utf-8')
        self.assertEqual(expected, mapped_params)

    def test_unknown_params(self):
        user_params = {'Foo': 'wibble'}
        self.assertRaises(exception.UnknownUserParameter,
                          self.new_parameters,
                          'test',
                          params_schema,
                          user_params)

    def test_missing_params(self):
        user_params = {}
        self.assertRaises(exception.UserParameterMissing,
                          self.new_parameters,
                          'test',
                          params_schema,
                          user_params)

    def test_missing_attribute_params(self):
        params = {'Parameters': {'Foo': {'Type': 'String'},
                                 'NoAttr': 'No attribute.',
                                 'Bar': {'Type': 'Number', 'Default': '1'}}}
        self.assertRaises(exception.InvalidSchemaError,
                          self.new_parameters,
                          'test',
                          params)


class ParameterDefaultsTest(ParametersBase):
    scenarios = [
        ('type_list', dict(p_type='CommaDelimitedList',
                           value='1,1,1',
                           expected=[['4', '2'], ['7', '7'], ['1', '1', '1']],
                           param_default='7,7',
                           default='4,2')),
        ('type_number', dict(p_type='Number',
                             value=111,
                             expected=[42, 77, 111],
                             param_default=77,
                             default=42)),
        ('type_string', dict(p_type='String',
                             value='111',
                             expected=['42', '77', '111'],
                             param_default='77',
                             default='42')),
        ('type_json', dict(p_type='Json',
                           value={'1': '11'},
                           expected=[{'4': '2'}, {'7': '7'}, {'1': '11'}],
                           param_default={'7': '7'},
                           default={'4': '2'})),
        ('type_boolean1', dict(p_type='Boolean',
                               value=True,
                               expected=[False, False, True],
                               param_default=False,
                               default=False)),
        ('type_boolean2', dict(p_type='Boolean',
                               value=False,
                               expected=[False, True, False],
                               param_default=True,
                               default=False)),
        ('type_boolean3', dict(p_type='Boolean',
                               value=False,
                               expected=[True, False, False],
                               param_default=False,
                               default=True))]

    def test_use_expected_default(self):
        template = {'Parameters': {'a': {'Type': self.p_type,
                                         'Default': self.default}}}
        params = self.new_parameters('test_params', template)
        self.assertEqual(self.expected[0], params['a'])

        params = self.new_parameters('test_params', template,
                                     param_defaults={'a': self.param_default})
        self.assertEqual(self.expected[1], params['a'])

        params = self.new_parameters('test_params', template,
                                     {'a': self.value},
                                     param_defaults={'a': self.param_default})
        self.assertEqual(self.expected[2], params['a'])


class ParameterSchemaTest(common.HeatTestCase):

    def test_validate_schema_wrong_key(self):
        error = self.assertRaises(exception.InvalidSchemaError,
                                  parameters.Schema.from_dict, 'param_name',
                                  {"foo": "bar"})
        self.assertEqual("Invalid key 'foo' for parameter (param_name)",
                         str(error))

    def test_validate_schema_no_type(self):
        error = self.assertRaises(exception.InvalidSchemaError,
                                  parameters.Schema.from_dict,
                                  'broken',
                                  {"Description": "Hi!"})
        self.assertEqual("Missing parameter type for parameter: broken",
                         str(error))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import properties
from heat.engine import resource
from heat.engine import software_config_io as swc_io
from heat.engine import support
from heat.rpc import api as rpc_api


class SoftwareConfig(resource.Resource):
    """A resource for describing and storing software configuration.

    The software_configs API which backs this resource creates immutable
    configs, so any change to the template resource definition will result
    in a new config being created, and the old one being deleted.

    Configs can be defined in the same template which uses them, or they can
    be created in one stack, and passed to another stack via a parameter.

    A config resource can be referenced in other resource properties which
    are config-aware. This includes the properties OS::Nova::Server user_data,
    OS::Heat::SoftwareDeployment config and OS::Heat::MultipartMime parts
    config.

    Along with the config script itself, this resource can define schemas for
    inputs and outputs which the config script is expected to consume and
    produce. Inputs and outputs are optional and will map to concepts which
    are specific to the configuration tool being used.
    """

    support_status = support.SupportStatus(version='2014.1')

    default_client_name = 'heat'

    entity = 'software_configs'

    PROPERTIES = (
        GROUP, CONFIG,
        OPTIONS,
        INPUTS, OUTPUTS
    ) = (
        rpc_api.SOFTWARE_CONFIG_GROUP, rpc_api.SOFTWARE_CONFIG_CONFIG,
        rpc_api.SOFTWARE_CONFIG_OPTIONS,
        rpc_api.SOFTWARE_CONFIG_INPUTS, rpc_api.SOFTWARE_CONFIG_OUTPUTS,
    )

    ATTRIBUTES = (
        CONFIG_ATTR,
    ) = (
        'config',
    )

    properties_schema = {
        GROUP: properties.Schema(
            properties.Schema.STRING,
            _('Namespace to group this software config by when delivered to '
              'a server. This may imply what configuration tool is going to '
              'perform the configuration.'),
            default='Heat::Ungrouped'
        ),
        CONFIG: properties.Schema(
            properties.Schema.STRING,
            _('Configuration script or manifest which specifies what actual '
              'configuration is performed.'),
        ),
        OPTIONS: properties.Schema(
            properties.Schema.MAP,
            _('Map containing options specific to the configuration '
              'management tool used by this resource.'),
        ),
        INPUTS: properties.Schema(
            properties.Schema.LIST,
            _('Schema representing the inputs that this software config is '
              'expecting.'),
            schema=properties.Schema(properties.Schema.MAP,
                                     schema=swc_io.input_config_schema)
        ),
        OUTPUTS: properties.Schema(
            properties.Schema.LIST,
            _('Schema representing the outputs that this software config '
              'will produce.'),
            schema=properties.Schema(properties.Schema.MAP,
                                     schema=swc_io.output_config_schema)
        ),
    }

    attributes_schema = {
        CONFIG_ATTR: attributes.Schema(
            _("The config value of the software config."),
            type=attributes.Schema.STRING
        ),
    }

    def handle_create(self):
        props = dict(self.properties)
        props[rpc_api.SOFTWARE_CONFIG_NAME] = self.physical_resource_name()

        sc = self.rpc_client().create_software_config(self.context, **props)
        self.resource_id_set(sc[rpc_api.SOFTWARE_CONFIG_ID])

    def handle_delete(self):

        if self.resource_id is None:
            return

        with self.rpc_client().ignore_error_by_name('NotFound'):
            self.rpc_client().delete_software_config(
                self.context, self.resource_id)

    def _resolve_attribute(self, name):
        """Retrieve attributes of the SoftwareConfig resource.

        "config" returns the config value of the software config. If the
         software config does not exist, returns an empty string.
        """
        if name == self.CONFIG_ATTR and self.resource_id:
            with self.rpc_client().ignore_error_by_name('NotFound'):
                sc = self.rpc_client().show_software_config(
                    self.context, self.resource_id)
                return sc[rpc_api.SOFTWARE_CONFIG_CONFIG]


def resource_mapping():
    return {
        'OS::Heat::SoftwareConfig': SoftwareConfig,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints as constr
from heat.engine import parameters


class CfnParameters(parameters.Parameters):
    PSEUDO_PARAMETERS = (
        PARAM_STACK_ID, PARAM_STACK_NAME, PARAM_REGION
    ) = (
        'AWS::StackId', 'AWS::StackName', 'AWS::Region'
    )

    def _pseudo_parameters(self, stack_identifier):
        stack_id = (stack_identifier.arn()
                    if stack_identifier is not None else 'None')
        stack_name = stack_identifier and stack_identifier.stack_name

        yield parameters.Parameter(
            self.PARAM_STACK_ID,
            parameters.Schema(parameters.Schema.STRING, _('Stack ID'),
                              default=str(stack_id)))
        if stack_name:
            yield parameters.Parameter(
                self.PARAM_STACK_NAME,
                parameters.Schema(parameters.Schema.STRING, _('Stack Name'),
                                  default=stack_name))
            yield parameters.Parameter(
                self.PARAM_REGION,
                parameters.Schema(parameters.Schema.STRING,
                                  default='ap-southeast-1',
                                  constraints=[
                                      constr.AllowedValues(
                                          ['us-east-1',
                                           'us-west-1',
                                           'us-west-2',
                                           'sa-east-1',
                                           'eu-west-1',
                                           'ap-southeast-1',
                                           'ap-northeast-1'])]))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from urllib import parse

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine.clients.os import swift
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support

LOG = logging.getLogger(__name__)


class SwiftSignalFailure(exception.Error):
    def __init__(self, wait_cond):
        reasons = wait_cond.get_status_reason(wait_cond.STATUS_FAILURE)
        super(SwiftSignalFailure, self).__init__(';'.join(reasons))


class SwiftSignalTimeout(exception.Error):
    def __init__(self, wait_cond):
        reasons = wait_cond.get_status_reason(wait_cond.STATUS_SUCCESS)
        vals = {'len': len(reasons),
                'count': wait_cond.properties[wait_cond.COUNT]}
        if reasons:
            vals['reasons'] = ';'.join(reasons)
            message = (_('%(len)d of %(count)d received - %(reasons)s') % vals)
        else:
            message = (_('%(len)d of %(count)d received') % vals)
        super(SwiftSignalTimeout, self).__init__(message)


class SwiftSignalHandle(resource.Resource):
    """Resource for managing signals from Swift resources.

    This resource is same as WaitConditionHandle, but designed for using by
    Swift resources.
    """

    support_status = support.SupportStatus(version='2014.2')
    default_client_name = "swift"

    properties_schema = {}

    ATTRIBUTES = (
        TOKEN,
        ENDPOINT,
        CURL_CLI,
    ) = (
        'token',
        'endpoint',
        'curl_cli',
    )

    attributes_schema = {
        TOKEN: attributes.Schema(
            _('Tokens are not needed for Swift TempURLs. This attribute is '
              'being kept for compatibility with the '
              'OS::Heat::WaitConditionHandle resource.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.STRING
        ),
        ENDPOINT: attributes.Schema(
            _('Endpoint/url which can be used for signalling handle.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.STRING
        ),
        CURL_CLI: attributes.Schema(
            _('Convenience attribute, provides curl CLI command '
              'prefix, which can be used for signalling handle completion or '
              'failure. You can signal success by adding '
              '--data-binary \'{"status": "SUCCESS"}\' '
              ', or signal failure by adding '
              '--data-binary \'{"status": "FAILURE"}\'.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.STRING
        ),
    }

    def handle_create(self):
        cplugin = self.client_plugin()
        url = cplugin.get_signal_url(self.stack.id,
                                     self.physical_resource_name())
        self.data_set(self.ENDPOINT, url)
        self.resource_id_set(self.physical_resource_name())

    def _resolve_attribute(self, key):
        if self.resource_id:
            if key == self.TOKEN:
                return ''  # HeatWaitConditionHandle compatibility
            elif key == self.ENDPOINT:
                return self.data().get(self.ENDPOINT)
            elif key == self.CURL_CLI:
                return ("curl -i -X PUT '%s'" %
                        self.data().get(self.ENDPOINT))

    def handle_delete(self):
        cplugin = self.client_plugin()
        client = cplugin.client()

        # Delete all versioned objects
        while True:
            try:
                client.delete_object(self.stack.id,
                                     self.physical_resource_name())
            except Exception as exc:
                cplugin.ignore_not_found(exc)
                break

        # Delete the container if it is empty
        try:
            client.delete_container(self.stack.id)
        except Exception as exc:
            if cplugin.is_not_found(exc) or cplugin.is_conflict(exc):
                pass
            else:
                raise

        self.data_delete(self.ENDPOINT)

    def get_reference_id(self):
        return self.data().get(self.ENDPOINT)


class SwiftSignal(resource.Resource):
    """Resource for handling signals received by SwiftSignalHandle.

    This resource handles signals received by SwiftSignalHandle and
    is same as WaitCondition resource.
    """

    support_status = support.SupportStatus(version='2014.2')
    default_client_name = "swift"

    PROPERTIES = (HANDLE, TIMEOUT, COUNT,) = ('handle', 'timeout', 'count',)

    properties_schema = {
        HANDLE: properties.Schema(
            properties.Schema.STRING,
            required=True,
            description=_('URL of TempURL where resource will signal '
                          'completion and optionally upload data.')
        ),
        TIMEOUT: properties.Schema(
            properties.Schema.NUMBER,
            description=_('The maximum number of seconds to wait for the '
                          'resource to signal completion. Once the timeout '
                          'is reached, creation of the signal resource will '
                          'fail.'),
            required=True,
            constraints=[
                constraints.Range(1, 43200),
            ]
        ),
        COUNT: properties.Schema(
            properties.Schema.INTEGER,
            description=_('The number of success signals that must be '
                          'received before the stack creation process '
                          'continues.'),
            default=1,
            constraints=[
                constraints.Range(1, 1000),
            ]
        )
    }

    ATTRIBUTES = (DATA) = 'data'

    attributes_schema = {
        DATA: attributes.Schema(
            _('JSON data that was uploaded via the SwiftSignalHandle.'),
            type=attributes.Schema.STRING
        )
    }

    WAIT_STATUSES = (
        STATUS_FAILURE,
        STATUS_SUCCESS,
    ) = (
        'FAILURE',
        'SUCCESS',
    )

    METADATA_KEYS = (
        DATA, REASON, STATUS, UNIQUE_ID
    ) = (
        'data', 'reason', 'status', 'id'
    )

    def __init__(self, name, json_snippet, stack):
        super(SwiftSignal, self).__init__(name, json_snippet, stack)
        self._obj_name = None
        self._url = None

    @property
    def url(self):
        if not self._url:
            self._url = parse.urlparse(self.properties[self.HANDLE])
        return self._url

    @property
    def obj_name(self):
        if not self._obj_name:
            self._obj_name = self.url.path.split('/')[4]
        return self._obj_name

    def _validate_handle_url(self):
        parts = self.url.path.split('/')
        msg = _('"%(url)s" is not a valid SwiftSignalHandle.  The %(part)s '
                'is invalid')
        cplugin = self.client_plugin()
        if not cplugin.is_valid_temp_url_path(self.url.path):
            raise ValueError(msg % {'url': self.url.path,
                                    'part': 'Swift TempURL path'})
        if not parts[3] == self.stack.id:
            raise ValueError(msg % {'url': self.url.path,
                                    'part': 'container name'})

    def handle_create(self):
        self._validate_handle_url()
        started_at = timeutils.utcnow()
        return started_at, float(self.properties[self.TIMEOUT])

    def get_signals(self):
        try:
            container = self.client().get_container(self.stack.id)
        except Exception as exc:
            self.client_plugin().ignore_not_found(exc)
            LOG.debug("Swift container %s was not found", self.stack.id)
            return []

        index = container[1]
        if not index:
            LOG.debug("Swift objects in container %s were not found",
                      self.stack.id)
            return []

        # Remove objects in that are for other handle resources, since
        # multiple SwiftSignalHandle resources in the same stack share
        # a container
        filtered = [obj for obj in index if self.obj_name in obj['name']]

        # Fetch objects from Swift and filter results
        obj_bodies = []
        for obj in filtered:
            try:
                signal = self.client().get_object(self.stack.id, obj['name'])
            except Exception as exc:
                self.client_plugin().ignore_not_found(exc)
                continue

            body = signal[1]
            if isinstance(body, bytes):
                body = body.decode()
            if body == swift.IN_PROGRESS:  # Ignore the initial object
                continue
            if body == "":
                obj_bodies.append({})
                continue
            try:
                obj_bodies.append(jsonutils.loads(body))
            except ValueError:
                raise exception.Error(_("Failed to parse JSON data: %s") %
                                      body)

        # Set default values on each signal
        signals = []
        signal_num = 1
        for signal in obj_bodies:

            # Remove previous signals with the same ID
            sig_id = self.UNIQUE_ID
            ids = [s.get(sig_id) for s in signals if sig_id in s]
            if ids and sig_id in signal and ids.count(signal[sig_id]) > 0:
                [signals.remove(s) for s in signals
                 if s.get(sig_id) == signal[sig_id]]

            # Make sure all fields are set, since all are optional
            signal.setdefault(self.DATA, None)
            unique_id = signal.setdefault(sig_id, signal_num)
            reason = 'Signal %s received' % unique_id
            signal.setdefault(self.REASON, reason)
            signal.setdefault(self.STATUS, self.STATUS_SUCCESS)

            signals.append(signal)
            signal_num += 1

        return signals

    def get_status(self):
        return [s[self.STATUS] for s in self.get_signals()]

    def get_status_reason(self, status):
        return [s[self.REASON]
                for s in self.get_signals()
                if s[self.STATUS] == status]

    def get_data(self):
        signals = self.get_signals()
        if not signals:
            return None
        data = {}
        for signal in signals:
            data[signal[self.UNIQUE_ID]] = signal[self.DATA]
        return data

    def check_create_complete(self, create_data):
        if timeutils.is_older_than(*create_data):
            raise SwiftSignalTimeout(self)

        statuses = self.get_status()
        if not statuses:
            return False

        for status in statuses:
            if status == self.STATUS_FAILURE:
                failure = SwiftSignalFailure(self)
                LOG.info('%(name)s Failed (%(failure)s)',
                         {'name': str(self), 'failure': str(failure)})
                raise failure
            elif status != self.STATUS_SUCCESS:
                raise exception.Error(_("Unknown status: %s") % status)

        if len(statuses) >= self.properties[self.COUNT]:
            LOG.info("%s Succeeded", str(self))
            return True
        return False

    def _resolve_attribute(self, key):
        if key == self.DATA:
            return str(jsonutils.dumps(self.get_data()))


def resource_mapping():
    return {'OS::Heat::SwiftSignal': SwiftSignal,
            'OS::Heat::SwiftSignalHandle': SwiftSignalHandle}



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import re

from heat.common.i18n import _

DNS_LABEL_MAX_LEN = 63
DNS_LABEL_REGEX = "[a-z0-9-]{1,%d}$" % DNS_LABEL_MAX_LEN
FQDN_MAX_LEN = 255


def is_prefix_subset(orig_prefixes, new_prefixes):
    """Check whether orig_prefixes is subset of new_prefixes.


    This takes valid prefix lists for orig_prefixes and new_prefixes,
    returns 'True', if orig_prefixes is subset of new_prefixes.
    """
    orig_set = netaddr.IPSet(orig_prefixes)
    new_set = netaddr.IPSet(new_prefixes)
    return orig_set.issubset(new_set)


def validate_dns_format(data):
    if not data:
        return
    trimmed = data if not data.endswith('.') else data[:-1]
    if len(trimmed) > FQDN_MAX_LEN:
        raise ValueError(
            _("'%(data)s' exceeds the %(max_len)s character FQDN limit") % {
                'data': trimmed,
                'max_len': FQDN_MAX_LEN})
    names = trimmed.split('.')
    for name in names:
        if not name:
            raise ValueError(_("Encountered an empty component."))
        if name.endswith('-') or name.startswith('-'):
            raise ValueError(
                _("Name '%s' must not start or end with a hyphen.") % name)
        if not re.match(DNS_LABEL_REGEX, name):
            raise ValueError(
                _("Name '%(name)s' must be 1-%(max_len)s characters long, "
                  "each of which can only be alphanumeric or "
                  "a hyphen.") % {'name': name,
                                  'max_len': DNS_LABEL_MAX_LEN})
    # RFC 1123 hints that a Top Level Domain(TLD) can't be all numeric.
    # Last part is a TLD, if it's a FQDN.
    if (data.endswith('.') and len(names) > 1
            and re.match("^[0-9]+$", names[-1])):
        raise ValueError(_("TLD '%s' must not be all numeric.") % names[-1])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.heat import scaling_policy as heat_sp
from heat.scaling import scalingutil as sc_util


class AWSScalingPolicy(heat_sp.AutoScalingPolicy):
    PROPERTIES = (
        AUTO_SCALING_GROUP_NAME, SCALING_ADJUSTMENT, ADJUSTMENT_TYPE,
        COOLDOWN, MIN_ADJUSTMENT_STEP,
    ) = (
        'AutoScalingGroupName', 'ScalingAdjustment', 'AdjustmentType',
        'Cooldown', 'MinAdjustmentStep',
    )

    ATTRIBUTES = (
        ALARM_URL,
    ) = (
        'AlarmUrl',
    )

    properties_schema = {
        AUTO_SCALING_GROUP_NAME: properties.Schema(
            properties.Schema.STRING,
            _('AutoScaling group name to apply policy to.'),
            required=True
        ),
        SCALING_ADJUSTMENT: properties.Schema(
            properties.Schema.INTEGER,
            _('Size of adjustment.'),
            required=True,
            update_allowed=True
        ),
        ADJUSTMENT_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Type of adjustment (absolute or percentage).'),
            required=True,
            constraints=[
                constraints.AllowedValues(
                    [sc_util.CFN_CHANGE_IN_CAPACITY,
                     sc_util.CFN_EXACT_CAPACITY,
                     sc_util.CFN_PERCENT_CHANGE_IN_CAPACITY]),
            ],
            update_allowed=True
        ),
        COOLDOWN: properties.Schema(
            properties.Schema.INTEGER,
            _('Cooldown period, in seconds.'),
            update_allowed=True
        ),
        MIN_ADJUSTMENT_STEP: properties.Schema(
            properties.Schema.INTEGER,
            _('Minimum number of resources that are added or removed '
              'when the AutoScaling group scales up or down. This can '
              'be used only when specifying PercentChangeInCapacity '
              'for the AdjustmentType property.'),
            constraints=[
                constraints.Range(
                    min=0,
                ),
            ],
            update_allowed=True
        ),

    }

    attributes_schema = {
        ALARM_URL: attributes.Schema(
            _("A signed url to handle the alarm. (Heat extension)."),
            type=attributes.Schema.STRING,
            cache_mode=attributes.Schema.CACHE_NONE
        ),
    }

    def _validate_min_adjustment_step(self):
        adjustment_type = self.properties.get(self.ADJUSTMENT_TYPE)
        adjustment_step = self.properties.get(self.MIN_ADJUSTMENT_STEP)
        if (adjustment_type != sc_util.CFN_PERCENT_CHANGE_IN_CAPACITY
                and adjustment_step is not None):
            raise exception.ResourcePropertyValueDependency(
                prop1=self.MIN_ADJUSTMENT_STEP,
                prop2=self.ADJUSTMENT_TYPE,
                value=sc_util.CFN_PERCENT_CHANGE_IN_CAPACITY)

    def get_reference_id(self):
        if self.resource_id is not None:
            return str(self._get_ec2_signed_url(never_expire=True))
        else:
            return str(self.name)


def resource_mapping():
    return {
        'AWS::AutoScaling::ScalingPolicy': AWSScalingPolicy,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock
import zoneinfo

from testtools import matchers

from heat.engine.clients.os import swift
from heat.tests import common
from heat.tests import utils


class SwiftClientPluginTestCase(common.HeatTestCase):
    def setUp(self):
        super(SwiftClientPluginTestCase, self).setUp()
        self.swift_client = mock.Mock()
        self.context = utils.dummy_context()
        self.context.project_id = "demo"
        c = self.context.clients
        self.swift_plugin = c.client_plugin('swift')
        self.swift_plugin.client = lambda: self.swift_client


class SwiftUtilsTest(SwiftClientPluginTestCase):

    def test_is_valid_temp_url_path(self):

        valids = [
            "/v1/AUTH_demo/c/o",
            "/v1/AUTH_demo/c/o/",
            "/v1/TEST_demo/c/o",
            "/v1/AUTH_demo/c/pseudo_folder/o",
        ]
        for url in valids:
            self.assertTrue(self.swift_plugin.is_valid_temp_url_path(url))

        invalids = [
            "/v2/AUTH_demo/c/o",
            "/v1/AUTH_demo/c//",
            "/v1/AUTH_demo/c/",
            "/AUTH_demo/c//",
            "//AUTH_demo/c/o",
            "//v1/AUTH_demo/c/o",
            "/v1/AUTH_demo/o",
            "/v1/AUTH_demo//o",
            "/v1/AUTH_d3mo//o",
            "/v1//c/o",
            "/v1/c/o",
        ]
        for url in invalids:
            self.assertFalse(self.swift_plugin.is_valid_temp_url_path(url))

    def test_get_temp_url(self):
        self.swift_client.url = ("http://fake-host.com:8080/v1/"
                                 "AUTH_demo")
        self.swift_client.head_account = mock.Mock(return_value={
            'x-account-meta-temp-url-key': '123456'})
        self.swift_client.post_account = mock.Mock()

        container_name = '1234'  # from stack.id
        stack_name = 'test'
        handle_name = 'foo'
        obj_name = '%s-%s' % (stack_name, handle_name)
        url = self.swift_plugin.get_temp_url(container_name, obj_name)
        self.assertFalse(self.swift_client.post_account.called)
        regexp = ("http://fake-host.com:8080/v1/AUTH_demo/%s"
                  r"/%s\?temp_url_sig=[0-9a-f]{40,64}&"
                  "temp_url_expires=[0-9]{10}" %
                  (container_name, obj_name))
        self.assertThat(url, matchers.MatchesRegex(regexp))

        timeout = int(url.split('=')[-1])
        self.assertLess(timeout, swift.MAX_EPOCH)

    def test_get_temp_url_no_account_key(self):
        self.swift_client.url = ("http://fake-host.com:8080/v1/"
                                 "AUTH_demo")
        head_account = {}

        def post_account(data):
            head_account.update(data)

        self.swift_client.head_account = mock.Mock(return_value=head_account)
        self.swift_client.post_account = post_account

        container_name = '1234'  # from stack.id
        stack_name = 'test'
        handle_name = 'foo'
        obj_name = '%s-%s' % (stack_name, handle_name)

        self.assertNotIn('x-account-meta-temp-url-key', head_account)
        self.swift_plugin.get_temp_url(container_name, obj_name)
        self.assertIn('x-account-meta-temp-url-key', head_account)

    def test_get_signal_url(self):
        self.swift_client.url = ("http://fake-host.com:8080/v1/"
                                 "AUTH_demo")
        self.swift_client.head_account = mock.Mock(return_value={
            'x-account-meta-temp-url-key': '123456'})
        self.swift_client.post_account = mock.Mock()

        container_name = '1234'  # from stack.id
        stack_name = 'test'
        handle_name = 'foo'
        obj_name = '%s-%s' % (stack_name, handle_name)
        url = self.swift_plugin.get_signal_url(container_name, obj_name)
        self.assertTrue(self.swift_client.put_container.called)
        self.assertTrue(self.swift_client.put_object.called)
        regexp = ("http://fake-host.com:8080/v1/AUTH_demo/%s"
                  r"/%s\?temp_url_sig=[0-9a-f]{40,64}&"
                  "temp_url_expires=[0-9]{10}" %
                  (container_name, obj_name))
        self.assertThat(url, matchers.MatchesRegex(regexp))

    def test_parse_last_modified(self):
        self.assertIsNone(self.swift_plugin.parse_last_modified(None))
        tz = zoneinfo.ZoneInfo('GMT')
        now = datetime.datetime(
            2015, 2, 5, 1, 4, 40, 0, tz)
        now_naive = datetime.datetime(
            2015, 2, 5, 1, 4, 40, 0)
        last_modified = now.strftime('%a, %d %b %Y %H:%M:%S %Z')
        self.assertEqual('Thu, 05 Feb 2015 01:04:40 GMT', last_modified)
        self.assertEqual(
            now_naive,
            self.swift_plugin.parse_last_modified(last_modified))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.clients import client_plugin


class DefaultClientPlugin(client_plugin.ClientPlugin):
    """A ClientPlugin that has no client.

    This is provided so that Resource can make use of the is_not_found() and
    is_conflict() methods even if the resource plugin has not specified a
    client plugin.
    """
    def _create(self, version=None):
        return super(DefaultClientPlugin, self)._create(version)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import re
import sys

from oslo_log import log

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import plugin_manager
from heat.engine import support

LOG = log.getLogger(__name__)


class HeatCustomGuidelines(object):

    _RULES = ['resource_descriptions', 'trailing_spaces']

    def __init__(self, exclude):
        self.error_count = 0
        self.resources_classes = []
        all_resources = _load_all_resources()
        for resource_type in all_resources:
            for rsrc_cls in all_resources[resource_type]:
                module = rsrc_cls.__module__
                # Skip hidden resources check guidelines
                if rsrc_cls.support_status.status == support.HIDDEN:
                    continue
                # Skip resources, which defined as template resource in
                # environment or cotrib resource
                if module in ('heat.engine.resources.template_resource',
                              'heat.engine.plugins'):
                    continue
                # Skip manually excluded folders
                path = module.replace('.', '/')
                if any(path.startswith(excl_path) for excl_path in exclude):
                    continue
                self.resources_classes.append(rsrc_cls)

    def run_check(self):
        print(_('Heat custom guidelines check started.'))
        for rule in self._RULES:
            getattr(self, 'check_%s' % rule)()
        if self.error_count > 0:
            print(_('Heat custom guidelines check failed - '
                    'found %s errors.') % self.error_count)
            sys.exit(1)
        else:
            print(_('Heat custom guidelines check succeeded.'))

    def check_resource_descriptions(self):
        for cls in self.resources_classes:
            # check resource's description
            self._check_resource_description(cls)
            # check properties' descriptions
            self._check_resource_schemas(cls, cls.properties_schema,
                                         'property')
            # check attributes' descriptions
            self._check_resource_schemas(cls, cls.attributes_schema,
                                         'attribute')
            # check methods descriptions
            self._check_resource_methods(cls)

    def _check_resource_description(self, resource):
        description = resource.__doc__
        if resource.support_status.status not in (support.SUPPORTED,
                                                  support.UNSUPPORTED):
            return
        kwargs = {'path': resource.__module__, 'details': resource.__name__}
        if not description:
            kwargs.update({'message': _("Resource description missing, "
                                        "should add resource description "
                                        "about resource's purpose")})
            self.print_guideline_error(**kwargs)
            return

        doclines = [key.strip() for key in description.split('\n')]
        if len(doclines) == 1 or (len(doclines) == 2 and doclines[-1] == ''):
            kwargs.update({'message': _("Resource description missing, "
                                        "should add resource description "
                                        "about resource's purpose")})
            self.print_guideline_error(**kwargs)
            return

        self._check_description_summary(doclines[0], kwargs, 'resource')
        self._check_description_details(doclines, kwargs, 'resource')

    def _check_resource_schemas(self, resource, schema, schema_name,
                                error_path=None):
        for key, value in schema.items():
            if error_path is None:
                error_path = [resource.__name__, key]
            else:
                error_path.append(key)
            # need to check sub-schema of current schema, if exists
            if (hasattr(value, 'schema') and
                    getattr(value, 'schema') is not None):
                self._check_resource_schemas(resource, value.schema,
                                             schema_name, error_path)
            description = value.description
            kwargs = {'path': resource.__module__, 'details': error_path}
            if description is None:
                if (value.support_status.status == support.SUPPORTED and
                        not isinstance(value.schema,
                                       constraints.AnyIndexDict) and
                        not isinstance(schema, constraints.AnyIndexDict)):
                    kwargs.update({'message': _("%s description "
                                                "missing, need to add "
                                                "description about property's "
                                                "purpose") % schema_name})
                    self.print_guideline_error(**kwargs)
                error_path.pop()
                continue
            self._check_description_summary(description, kwargs, schema_name)
            error_path.pop()

    def _check_resource_methods(self, resource):
        for method in resource.__dict__.values():
            # need to skip non-functions attributes
            if not callable(method):
                continue
            description = method.__doc__
            if not description:
                continue
            if method.__name__.startswith('__'):
                continue
            doclines = [key.strip() for key in description.split('\n')]
            kwargs = {'path': resource.__module__,
                      'details': [resource.__name__, method.__name__]}

            self._check_description_summary(doclines[0], kwargs, 'method')

            if len(doclines) == 2:
                kwargs.update({'message': _('Method description summary '
                                            'should be in one line')})
                self.print_guideline_error(**kwargs)
                continue

            if len(doclines) > 1:
                self._check_description_details(doclines, kwargs, 'method')

    def check_trailing_spaces(self):
        for cls in self.resources_classes:
            try:
                cls_file = open(cls.__module__.replace('.', '/') + '.py')
            except IOError as ex:
                LOG.warning('Cannot perform trailing spaces check on '
                            'resource module: %s', str(ex))
                continue
            lines = [line.strip() for line in cls_file.readlines()]
            idx = 0
            kwargs = {'path': cls.__module__}
            while idx < len(lines):
                if ('properties_schema' in lines[idx] or
                        'attributes_schema' in lines[idx]):
                    level = len(re.findall(r'(\{|\()', lines[idx]))
                    level -= len(re.findall(r'(\}|\))', lines[idx]))
                    idx += 1
                    while level != 0:
                        level += len(re.findall(r'(\{|\()', lines[idx]))
                        level -= len(re.findall(r'(\}|\))', lines[idx]))
                        if re.search("^((\'|\") )", lines[idx]):
                            kwargs.update(
                                {'details': 'line %s' % idx,
                                 'message': _('Trailing whitespace should '
                                              'be on previous line'),
                                 'snippet': lines[idx]})
                            self.print_guideline_error(**kwargs)
                        elif (re.search("(\\S(\'|\"))$", lines[idx - 1]) and
                              re.search("^((\'|\")\\S)", lines[idx])):
                            kwargs.update(
                                {'details': 'line %s' % (idx - 1),
                                 'message': _('Omitted whitespace at the '
                                              'end of the line'),
                                 'snippet': lines[idx - 1]})
                            self.print_guideline_error(**kwargs)
                        idx += 1
                idx += 1

    def _check_description_summary(self, description, error_kwargs,
                                   error_key):
        if re.search("^[a-z]", description):
            error_kwargs.update(
                {'message': _('%s description summary should start '
                              'with uppercase letter') % error_key.title(),
                 'snippet': description})
            self.print_guideline_error(**error_kwargs)
        if not (description.endswith('.') or description.endswith('.)')):
            error_kwargs.update(
                {'message': _('%s description summary omitted '
                              'terminator at the end') % error_key.title(),
                 'snippet': description})
            self.print_guideline_error(**error_kwargs)
        if re.search(r"\s{2,}", description):
            error_kwargs.update(
                {'message': _('%s description contains double or more '
                              'whitespaces') % error_key.title(),
                 'snippet': description})
            self.print_guideline_error(**error_kwargs)

    def _check_description_details(self, doclines, error_kwargs,
                                   error_key):
        if re.search(r"\S", doclines[1]):
            error_kwargs.update(
                {'message': _('%s description summary and '
                              'main resource description should be '
                              'separated by blank line') % error_key.title(),
                 'snippet': doclines[0]})
            self.print_guideline_error(**error_kwargs)

        if re.search("^[a-z]", doclines[2]):
            error_kwargs.update(
                {'message': _('%s description should start '
                              'with with uppercase '
                              'letter') % error_key.title(),
                 'snippet': doclines[2]})
            self.print_guideline_error(**error_kwargs)

        if doclines[-1] != '':
            error_kwargs.update(
                {'message': _('%s description multistring '
                              'should have singly closing quotes at '
                              'the next line') % error_key.title(),
                 'snippet': doclines[-1]})
            self.print_guideline_error(**error_kwargs)

        params = False
        for line in doclines[1:]:
                if re.search(r"\s{2,}", line):
                    error_kwargs.update(
                        {'message': _('%s description '
                                      'contains double or more '
                                      'whitespaces') % error_key.title(),
                         'snippet': line})
                    self.print_guideline_error(**error_kwargs)
                if re.search("^(:param|:type|:returns|:rtype|:raises)",
                             line):
                    params = True
        if not params and not (doclines[-2].endswith('.') or
                               doclines[-2].endswith('.)')):
            error_kwargs.update(
                {'message': _('%s description omitted '
                              'terminator at the end') % error_key.title(),
                 'snippet': doclines[-2]})
            self.print_guideline_error(**error_kwargs)

    def print_guideline_error(self, path, details, message, snippet=None):
        if isinstance(details, list):
            details = '.'.join(details)
        msg = _('ERROR (in %(path)s: %(details)s): %(message)s') % {
            'message': message,
            'path': path.replace('.', '/'),
            'details': details
        }
        if snippet is not None:
            msg = _('%(msg)s\n    (Error snippet): %(snippet)s') % {
                'msg': msg,
                'snippet': '%s...' % snippet[:79]
            }
        print(msg)
        self.error_count += 1


def _load_all_resources():
    manager = plugin_manager.PluginManager('heat.engine.resources')
    resource_mapping = plugin_manager.PluginMapping('resource')
    res_plugin_mappings = resource_mapping.load_all(manager)

    all_resources = {}
    for mapping in res_plugin_mappings:
        name, cls = mapping
        if all_resources.get(name) is not None:
            all_resources[name].append(cls)
        else:
            all_resources[name] = [cls]
    return all_resources


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--exclude', '-e', metavar='<FOLDER>',
                        nargs='+',
                        help=_('Exclude specified paths from checking.'))
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    guidelines = HeatCustomGuidelines(args.exclude or [])
    guidelines.run_check()



#
# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Client side of the heat engine RPC API."""

from oslo_utils import excutils
from oslo_utils import reflection

from heat.common import messaging
from heat.rpc import api as rpc_api


class EngineClient(object):
    """Client side of the heat engine rpc API.

    API version history::

        1.0 - Initial version.
        1.1 - Add support_status argument to list_resource_types()
        1.4 - Add support for service list
        1.9 - Add template_type option to generate_template()
        1.10 - Add support for software config list
        1.11 - Add support for template versions list
        1.12 - Add with_detail option for stack resources list
        1.13 - Add support for template functions list
        1.14 - Add cancel_with_rollback option to stack_cancel_update
        1.15 - Add preview_update_stack() call
        1.16 - Adds version, type_name to list_resource_types()
        1.17 - Add files to validate_template
        1.18 - Add show_nested to validate_template
        1.19 - Add show_output and list_outputs for returning stack outputs
        1.20 - Add resolve_outputs to stack show
        1.21 - Add deployment_id to create_software_deployment
        1.22 - Add support for stack export
        1.23 - Add environment_files to create/update/preview/validate
        1.24 - Adds ignorable_errors to validate_template
        1.25 - list_stack_resource filter update
        1.26 - Add mark_unhealthy
        1.27 - Add check_software_deployment
        1.28 - Add get_environment call
        1.29 - Add template_id to create_stack/update_stack
        1.30 - Add possibility to resource_type_* return descriptions
        1.31 - Add nested_depth to list_events, when nested_depth is specified
               add root_stack_id to response
        1.32 - Add get_files call
        1.33 - Remove tenant_safe from list_stacks, count_stacks
               and list_software_configs
        1.34 - Add migrate_convergence_1 call
        1.35 - Add with_condition to list_template_functions
        1.36 - Add files_container to create/update/preview/validate
    """

    BASE_RPC_API_VERSION = '1.0'

    def __init__(self):
        self._client = messaging.get_rpc_client(
            topic=rpc_api.ENGINE_TOPIC,
            version=self.BASE_RPC_API_VERSION)

    @staticmethod
    def make_msg(method, **kwargs):
        return method, kwargs

    def call(self, ctxt, msg, version=None, timeout=None):
        method, kwargs = msg

        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client

        if timeout is not None:
            client = client.prepare(timeout=timeout)

        return client.call(ctxt, method, **kwargs)

    def cast(self, ctxt, msg, version=None):
        method, kwargs = msg
        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client
        return client.cast(ctxt, method, **kwargs)

    def local_error_name(self, error):
        """Returns the name of the error with any _Remote postfix removed.

        :param error: Remote raised error to derive the name from.
        """
        error_name = reflection.get_class_name(error, fully_qualified=False)
        return error_name.split('_Remote')[0]

    def ignore_error_by_name(self, name):
        """Returns a context manager that filters exceptions with a given name.

        :param name: Name to compare the local exception name to.
        """
        def error_name_matches(err):
            return self.local_error_name(err) == name

        return excutils.exception_filter(error_name_matches)

    def identify_stack(self, ctxt, stack_name):
        """Returns the full stack identifier for a single, live stack.

        :param ctxt: RPC context.
        :param stack_name: Name of the stack you want to see,
                           or None to see all
        """
        return self.call(ctxt, self.make_msg('identify_stack',
                                             stack_name=stack_name))

    def list_stacks(self, ctxt, limit=None, marker=None, sort_keys=None,
                    sort_dir=None, filters=None,
                    show_deleted=False, show_nested=False, show_hidden=False,
                    tags=None, tags_any=None, not_tags=None,
                    not_tags_any=None):
        """Returns attributes of all stacks.

        It supports pagination (``limit`` and ``marker``), sorting
        (``sort_keys`` and ``sort_dir``) and filtering (``filters``) of the
        results.

        :param ctxt: RPC context.
        :param limit: the number of stacks to list (integer or string)
        :param marker: the ID of the last item in the previous page
        :param sort_keys: an array of fields used to sort the list
        :param sort_dir: the direction of the sort ('asc' or 'desc')
        :param filters: a dict with attribute:value to filter the list
        :param show_deleted: if true, show soft-deleted stacks
        :param show_nested: if true, show nested stacks
        :param show_hidden: if true, show hidden stacks
        :param tags: show stacks containing these tags. If multiple tags
            are passed, they will be combined using the boolean AND expression
        :param tags_any: show stacks containing these tags. If multiple tags
            are passed, they will be combined using the boolean OR expression
        :param not_tags: show stacks not containing these tags. If multiple
            tags are passed, they will be combined using the boolean AND
            expression
        :param not_tags_any: show stacks not containing these tags. If
            multiple tags are passed, they will be combined using the boolean
            OR expression
        :returns: a list of stacks
        """
        return self.call(ctxt,
                         self.make_msg('list_stacks', limit=limit,
                                       sort_keys=sort_keys, marker=marker,
                                       sort_dir=sort_dir, filters=filters,
                                       show_deleted=show_deleted,
                                       show_nested=show_nested,
                                       show_hidden=show_hidden,
                                       tags=tags, tags_any=tags_any,
                                       not_tags=not_tags,
                                       not_tags_any=not_tags_any),
                         version='1.33')

    def count_stacks(self, ctxt, filters=None,
                     show_deleted=False, show_nested=False, show_hidden=False,
                     tags=None, tags_any=None, not_tags=None,
                     not_tags_any=None):
        """Returns the number of stacks that match the given filters.

        :param ctxt: RPC context.
        :param filters: a dict of ATTR:VALUE to match against stacks
        :param show_deleted: if true, count will include the deleted stacks
        :param show_nested: if true, count will include nested stacks
        :param show_hidden: if true, count will include hidden stacks
        :param tags: count stacks containing these tags. If multiple tags are
            passed, they will be combined using the boolean AND expression
        :param tags_any: count stacks containing these tags. If multiple tags
            are passed, they will be combined using the boolean OR expression
        :param not_tags: count stacks not containing these tags. If multiple
            tags are passed, they will be combined using the boolean AND
            expression
        :param not_tags_any: count stacks not containing these tags. If
            multiple tags are passed, they will be combined using the boolean
            OR expression
        :returns: an integer representing the number of matched stacks
        """
        return self.call(ctxt, self.make_msg('count_stacks',
                                             filters=filters,
                                             show_deleted=show_deleted,
                                             show_nested=show_nested,
                                             show_hidden=show_hidden,
                                             tags=tags,
                                             tags_any=tags_any,
                                             not_tags=not_tags,
                                             not_tags_any=not_tags_any),
                         version='1.33')

    def show_stack(self, ctxt, stack_identity, resolve_outputs=True):
        """Returns detailed information about one or all stacks.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack you want to show, or None to
                               show all
        :param resolve_outputs: If True, stack outputs will be resolved
        """
        return self.call(ctxt, self.make_msg('show_stack',
                                             stack_identity=stack_identity,
                                             resolve_outputs=resolve_outputs),
                         version='1.20')

    def preview_stack(self, ctxt, stack_name, template, params, files,
                      args, environment_files=None, files_container=None):
        """Simulates a new stack using the provided template.

        Note that at this stage the template has already been fetched from the
        heat-api process if using a template-url.

        :param ctxt: RPC context.
        :param stack_name: Name of the stack you want to create.
        :param template: Template of stack you want to create.
        :param params: Stack Input Params/Environment
        :param files: files referenced from the environment.
        :param args: Request parameters/args passed from API
        :param environment_files: optional ordered list of environment file
               names included in the files dict
        :type  environment_files: list or None
        :param files_container: name of swift container
        """
        return self.call(ctxt,
                         self.make_msg('preview_stack', stack_name=stack_name,
                                       template=template,
                                       params=params, files=files,
                                       environment_files=environment_files,
                                       files_container=files_container,
                                       args=args),
                         version='1.36')

    def create_stack(self, ctxt, stack_name, template, params, files,
                     args, environment_files=None, files_container=None):
        """Creates a new stack using the template provided.

        Note that at this stage the template has already been fetched from the
        heat-api process if using a template-url.

        :param ctxt: RPC context.
        :param stack_name: Name of the stack you want to create.
        :param template: Template of stack you want to create.
        :param params: Stack Input Params/Environment
        :param files: files referenced from the environment.
        :param args: Request parameters/args passed from API
        :param environment_files: optional ordered list of environment file
               names included in the files dict
        :type  environment_files: list or None
        :param files_container: name of swift container
        """
        return self._create_stack(ctxt, stack_name, template, params, files,
                                  args, environment_files=environment_files,
                                  files_container=files_container)

    def _create_stack(self, ctxt, stack_name, template, params, files,
                      args, environment_files=None, files_container=None,
                      owner_id=None, nested_depth=0, user_creds_id=None,
                      stack_user_project_id=None, parent_resource_name=None,
                      template_id=None):
        """Internal interface for engine-to-engine communication via RPC.

        Allows some additional options which should not be exposed to users via
        the API:

        :param owner_id: parent stack ID for nested stacks
        :param nested_depth: nested depth for nested stacks
        :param user_creds_id: user_creds record for nested stack
        :param stack_user_project_id: stack user project for nested stack
        :param parent_resource_name: the parent resource name
        :param template_id: the ID of a pre-stored template in the DB
        """
        return self.call(
            ctxt, self.make_msg('create_stack', stack_name=stack_name,
                                template=template,
                                params=params, files=files,
                                environment_files=environment_files,
                                files_container=files_container,
                                args=args, owner_id=owner_id,
                                nested_depth=nested_depth,
                                user_creds_id=user_creds_id,
                                stack_user_project_id=stack_user_project_id,
                                parent_resource_name=parent_resource_name,
                                template_id=template_id),
            version='1.36')

    def update_stack(self, ctxt, stack_identity, template, params,
                     files, args, environment_files=None,
                     files_container=None):
        """Updates an existing stack based on the provided template and params.

        Note that at this stage the template has already been fetched from the
        heat-api process if using a template-url.

        :param ctxt: RPC context.
        :param stack_name: Name of the stack you want to create.
        :param template: Template of stack you want to create.
        :param params: Stack Input Params/Environment
        :param files: files referenced from the environment.
        :param args: Request parameters/args passed from API
        :param environment_files: optional ordered list of environment file
               names included in the files dict
        :type  environment_files: list or None
        :param files_container: name of swift container
        """
        return self._update_stack(ctxt, stack_identity, template, params,
                                  files, args,
                                  environment_files=environment_files,
                                  files_container=files_container)

    def _update_stack(self, ctxt, stack_identity, template, params,
                      files, args, environment_files=None,
                      files_container=None, template_id=None):
        """Internal interface for engine-to-engine communication via RPC.

        Allows an additional option which should not be exposed to users via
        the API:

        :param template_id: the ID of a pre-stored template in the DB
        """
        return self.call(ctxt,
                         self.make_msg('update_stack',
                                       stack_identity=stack_identity,
                                       template=template,
                                       params=params,
                                       files=files,
                                       environment_files=environment_files,
                                       files_container=files_container,
                                       args=args,
                                       template_id=template_id),
                         version='1.36')

    def preview_update_stack(self, ctxt, stack_identity, template, params,
                             files, args, environment_files=None,
                             files_container=None):
        """Returns the resources that would be changed in an update.

        Based on the provided template and parameters.

        Requires RPC version 1.15 or above.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack you wish to update.
        :param template: New template for the stack.
        :param params: Stack Input Params/Environment
        :param files: files referenced from the environment.
        :param args: Request parameters/args passed from API
        :param environment_files: optional ordered list of environment file
               names included in the files dict
        :type  environment_files: list or None
        :param files_container: name of swift container
        """
        return self.call(ctxt,
                         self.make_msg('preview_update_stack',
                                       stack_identity=stack_identity,
                                       template=template,
                                       params=params,
                                       files=files,
                                       environment_files=environment_files,
                                       files_container=files_container,
                                       args=args,
                                       ),
                         version='1.36')

    def validate_template(self, ctxt, template, params=None, files=None,
                          environment_files=None, files_container=None,
                          show_nested=False, ignorable_errors=None):
        """Uses the stack parser to check the validity of a template.

        :param ctxt: RPC context.
        :param template: Template of stack you want to create.
        :param params: Stack Input Params/Environment
        :param files: files referenced from the environment/template.
        :param environment_files: ordered list of environment file names
                                  included in the files dict
        :param files_container: name of swift container
        :param show_nested: if True nested templates will be validated
        :param ignorable_errors: List of error_code to be ignored as part of
                                 validation
        """
        return self.call(ctxt, self.make_msg(
            'validate_template',
            template=template,
            params=params,
            files=files,
            show_nested=show_nested,
            environment_files=environment_files,
            files_container=files_container,
            ignorable_errors=ignorable_errors),
            version='1.36')

    def authenticated_to_backend(self, ctxt):
        """Validate the credentials in the RPC context.

        Verify that the credentials in the RPC context are valid for the
        current cloud backend.

        :param ctxt: RPC context.
        """
        return self.call(ctxt, self.make_msg('authenticated_to_backend'))

    def get_template(self, ctxt, stack_identity):
        """Get the template.

        :param ctxt: RPC context.
        :param stack_name: Name of the stack you want to see.
        """
        return self.call(ctxt, self.make_msg('get_template',
                                             stack_identity=stack_identity))

    def get_environment(self, context, stack_identity):
        """Returns the environment for an existing stack.

        :param context: RPC context
        :param stack_identity: identifies the stack
        :rtype: dict
        """

        return self.call(context,
                         self.make_msg('get_environment',
                                       stack_identity=stack_identity),
                         version='1.28')

    def get_files(self, context, stack_identity):
        """Returns the files for an existing stack.

        :param context: RPC context
        :param stack_identity: identifies the stack
        :rtype: dict
        """

        return self.call(context,
                         self.make_msg('get_files',
                                       stack_identity=stack_identity),
                         version='1.32')

    def delete_stack(self, ctxt, stack_identity, cast=False):
        """Deletes a given stack.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack you want to delete.
        :param cast: cast the message instead of using call (default: False)

        You probably never want to use cast(). If you do, you'll never hear
        about any exceptions the call might raise.
        """
        rpc_method = self.cast if cast else self.call
        return rpc_method(ctxt,
                          self.make_msg('delete_stack',
                                        stack_identity=stack_identity))

    def abandon_stack(self, ctxt, stack_identity):
        """Deletes a given stack but resources would not be deleted.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack you want to abandon.
        """
        return self.call(ctxt,
                         self.make_msg('abandon_stack',
                                       stack_identity=stack_identity))

    def list_resource_types(self,
                            ctxt,
                            support_status=None,
                            type_name=None,
                            heat_version=None,
                            with_description=False):
        """Get a list of valid resource types.

        :param ctxt: RPC context.
        :param support_status: Support status of resource type
        :param type_name: Resource type's name (regular expression allowed)
        :param heat_version: Heat version
        :param with_description: Either return resource type description or not
        """
        return self.call(ctxt,
                         self.make_msg('list_resource_types',
                                       support_status=support_status,
                                       type_name=type_name,
                                       heat_version=heat_version,
                                       with_description=with_description),
                         version='1.30')

    def list_template_versions(self, ctxt):
        """Get a list of available template versions.

        :param ctxt: RPC context.
        """
        return self.call(ctxt, self.make_msg('list_template_versions'),
                         version='1.11')

    def list_template_functions(self, ctxt, template_version,
                                with_condition=False):
        """Get a list of available functions in a given template type.

        :param ctxt: RPC context
        :param template_version: template format/version tuple for which you
                                 want to get the list of functions.
        :param with_condition: return includes condition functions.
        """
        return self.call(ctxt,
                         self.make_msg('list_template_functions',
                                       template_version=template_version,
                                       with_condition=with_condition),
                         version='1.35')

    def resource_schema(self, ctxt, type_name, with_description=False):
        """Get the schema for a resource type.

        :param ctxt: RPC context.
        :param with_description: Return resource with description or not.
        """
        return self.call(ctxt,
                         self.make_msg('resource_schema',
                                       type_name=type_name,
                                       with_description=with_description),
                         version='1.30')

    def generate_template(self, ctxt, type_name, template_type='cfn'):
        """Generate a template based on the specified type.

        :param ctxt: RPC context.
        :param type_name: The resource type name to generate a template for.
        :param template_type: the template type to generate, cfn or hot.
        """
        return self.call(ctxt, self.make_msg('generate_template',
                                             type_name=type_name,
                                             template_type=template_type),
                         version='1.9')

    def list_events(self, ctxt, stack_identity, filters=None, limit=None,
                    marker=None, sort_keys=None, sort_dir=None,
                    nested_depth=None):
        """Lists all events associated with a given stack.

        It supports pagination (``limit`` and ``marker``),
        sorting (``sort_keys`` and ``sort_dir``) and filtering(filters)
        of the results.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack you want to get events for
        :param filters: a dict with attribute:value to filter the list
        :param limit: the number of events to list (integer or string)
        :param marker: the ID of the last event in the previous page
        :param sort_keys: an array of fields used to sort the list
        :param sort_dir: the direction of the sort ('asc' or 'desc').
        :param nested_depth: Levels of nested stacks to list events for.
        """
        return self.call(ctxt, self.make_msg('list_events',
                                             stack_identity=stack_identity,
                                             filters=filters,
                                             limit=limit,
                                             marker=marker,
                                             sort_keys=sort_keys,
                                             sort_dir=sort_dir,
                                             nested_depth=nested_depth),
                         version='1.31')

    def describe_stack_resource(self, ctxt, stack_identity, resource_name,
                                with_attr=False):
        """Get detailed resource information about a particular resource.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack.
        :param resource_name: the Resource.
        """
        return self.call(ctxt,
                         self.make_msg('describe_stack_resource',
                                       stack_identity=stack_identity,
                                       resource_name=resource_name,
                                       with_attr=with_attr),
                         version='1.2')

    def find_physical_resource(self, ctxt, physical_resource_id):
        """Return an identifier for the resource.

        :param ctxt: RPC context.
        :param physcial_resource_id: The physical resource ID to look up.
        """
        return self.call(ctxt,
                         self.make_msg(
                             'find_physical_resource',
                             physical_resource_id=physical_resource_id))

    def describe_stack_resources(self, ctxt, stack_identity, resource_name):
        """Get detailed resource information about one or more resources.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack.
        :param resource_name: the Resource.
        """
        return self.call(ctxt, self.make_msg('describe_stack_resources',
                                             stack_identity=stack_identity,
                                             resource_name=resource_name))

    def list_stack_resources(self, ctxt, stack_identity,
                             nested_depth=0, with_detail=False,
                             filters=None):
        """List the resources belonging to a stack.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack.
        :param nested_depth: Levels of nested stacks of which list resources.
        :param with_detail: show detail for resources in list.
        :param filters: a dict with attribute:value to search the resources
        """
        return self.call(ctxt,
                         self.make_msg('list_stack_resources',
                                       stack_identity=stack_identity,
                                       nested_depth=nested_depth,
                                       with_detail=with_detail,
                                       filters=filters),
                         version='1.25')

    def stack_suspend(self, ctxt, stack_identity):
        return self.call(ctxt, self.make_msg('stack_suspend',
                                             stack_identity=stack_identity))

    def stack_resume(self, ctxt, stack_identity):
        return self.call(ctxt, self.make_msg('stack_resume',
                                             stack_identity=stack_identity))

    def stack_check(self, ctxt, stack_identity):
        return self.call(ctxt, self.make_msg('stack_check',
                                             stack_identity=stack_identity))

    def stack_cancel_update(self, ctxt, stack_identity,
                            cancel_with_rollback=True):
        return self.call(ctxt,
                         self.make_msg(
                             'stack_cancel_update',
                             stack_identity=stack_identity,
                             cancel_with_rollback=cancel_with_rollback),
                         version='1.14')

    def resource_signal(self, ctxt, stack_identity, resource_name, details,
                        sync_call=False):
        """Generate an alarm on the resource.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack.
        :param resource_name: the Resource.
        :param details: the details of the signal.
        """
        return self.call(ctxt, self.make_msg('resource_signal',
                                             stack_identity=stack_identity,
                                             resource_name=resource_name,
                                             details=details,
                                             sync_call=sync_call),

                         version='1.3')

    def resource_mark_unhealthy(self, ctxt, stack_identity, resource_name,
                                mark_unhealthy, resource_status_reason=None):
        """Mark the resource as unhealthy or healthy.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack.
        :param resource_name: the Resource.
        :param mark_unhealthy: indicates whether the resource is unhealthy.
        :param resource_status_reason: reason for health change.
        """
        return self.call(
            ctxt,
            self.make_msg('resource_mark_unhealthy',
                          stack_identity=stack_identity,
                          resource_name=resource_name,
                          mark_unhealthy=mark_unhealthy,
                          resource_status_reason=resource_status_reason),
            version='1.26')

    def get_revision(self, ctxt):
        return self.call(ctxt, self.make_msg('get_revision'))

    def show_software_config(self, cnxt, config_id):
        return self.call(cnxt, self.make_msg('show_software_config',
                                             config_id=config_id))

    def list_software_configs(self, cnxt, limit=None, marker=None):
        return self.call(cnxt,
                         self.make_msg('list_software_configs',
                                       limit=limit,
                                       marker=marker),
                         version='1.33')

    def create_software_config(self, cnxt, group, name, config,
                               inputs=None, outputs=None, options=None):
        inputs = inputs or []
        outputs = outputs or []
        options = options or {}
        return self.call(cnxt, self.make_msg('create_software_config',
                                             group=group,
                                             name=name,
                                             config=config,
                                             inputs=inputs,
                                             outputs=outputs,
                                             options=options))

    def delete_software_config(self, cnxt, config_id):
        return self.call(cnxt, self.make_msg('delete_software_config',
                                             config_id=config_id))

    def list_software_deployments(self, cnxt, server_id=None):
        return self.call(cnxt, self.make_msg('list_software_deployments',
                                             server_id=server_id))

    def metadata_software_deployments(self, cnxt, server_id):
        return self.call(cnxt, self.make_msg('metadata_software_deployments',
                                             server_id=server_id))

    def show_software_deployment(self, cnxt, deployment_id):
        return self.call(cnxt, self.make_msg('show_software_deployment',
                                             deployment_id=deployment_id))

    def check_software_deployment(self, cnxt, deployment_id, timeout):
        return self.call(cnxt, self.make_msg('check_software_deployment',
                                             deployment_id=deployment_id,
                                             timeout=timeout),
                         timeout=timeout, version='1.27')

    def create_software_deployment(self, cnxt, server_id, config_id=None,
                                   input_values=None, action='INIT',
                                   status='COMPLETE', status_reason='',
                                   stack_user_project_id=None,
                                   deployment_id=None):
        input_values = input_values or {}
        return self.call(cnxt, self.make_msg(
            'create_software_deployment',
            server_id=server_id,
            config_id=config_id,
            deployment_id=deployment_id,
            input_values=input_values,
            action=action,
            status=status,
            status_reason=status_reason,
            stack_user_project_id=stack_user_project_id))

    def update_software_deployment(self, cnxt, deployment_id,
                                   config_id=None, input_values=None,
                                   output_values=None, action=None,
                                   status=None, status_reason=None,
                                   updated_at=None):
        return self.call(
            cnxt, self.make_msg('update_software_deployment',
                                deployment_id=deployment_id,
                                config_id=config_id,
                                input_values=input_values,
                                output_values=output_values,
                                action=action,
                                status=status,
                                status_reason=status_reason,
                                updated_at=updated_at),
            version='1.5')

    def delete_software_deployment(self, cnxt, deployment_id):
        return self.call(cnxt, self.make_msg('delete_software_deployment',
                                             deployment_id=deployment_id))

    def signal_software_deployment(self, cnxt, deployment_id, details,
                                   updated_at=None):
        return self.call(
            cnxt, self.make_msg('signal_software_deployment',
                                deployment_id=deployment_id,
                                details=details,
                                updated_at=updated_at),
            version='1.6')

    def stack_snapshot(self, ctxt, stack_identity, name):
        return self.call(ctxt, self.make_msg('stack_snapshot',
                                             stack_identity=stack_identity,
                                             name=name))

    def show_snapshot(self, cnxt, stack_identity, snapshot_id):
        return self.call(cnxt, self.make_msg('show_snapshot',
                                             stack_identity=stack_identity,
                                             snapshot_id=snapshot_id))

    def delete_snapshot(self, cnxt, stack_identity, snapshot_id):
        return self.call(cnxt, self.make_msg('delete_snapshot',
                                             stack_identity=stack_identity,
                                             snapshot_id=snapshot_id))

    def stack_list_snapshots(self, cnxt, stack_identity):
        return self.call(cnxt, self.make_msg('stack_list_snapshots',
                                             stack_identity=stack_identity))

    def stack_restore(self, cnxt, stack_identity, snapshot_id):
        return self.call(cnxt, self.make_msg('stack_restore',
                                             stack_identity=stack_identity,
                                             snapshot_id=snapshot_id))

    def list_services(self, cnxt):
        return self.call(cnxt, self.make_msg('list_services'), version='1.4')

    def list_outputs(self, cntx, stack_identity):
        return self.call(cntx, self.make_msg('list_outputs',
                                             stack_identity=stack_identity),
                         version='1.19')

    def show_output(self, cntx, stack_identity, output_key):
        return self.call(cntx, self.make_msg('show_output',
                                             stack_identity=stack_identity,
                                             output_key=output_key),
                         version='1.19')

    def export_stack(self, ctxt, stack_identity):
        """Exports the stack data in JSON format.

        :param ctxt: RPC context.
        :param stack_identity: Name of the stack you want to export.
        """
        return self.call(ctxt,
                         self.make_msg('export_stack',
                                       stack_identity=stack_identity),
                         version='1.22')

    def migrate_convergence_1(self, ctxt, stack_id):
        """Migrate the stack to convergence engine

        :param ctxt: RPC context
        :param stack_name: Name of the stack you want to migrate
        """
        return self.call(ctxt,
                         self.make_msg('migrate_convergence_1',
                                       stack_id=stack_id),
                         version='1.34')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from troveclient import exceptions as troveexc

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine.clients.os import trove
from heat.engine.resources.openstack.trove import cluster
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


stack_template = '''
heat_template_version: 2013-05-23

resources:

  cluster:
    type: OS::Trove::Cluster
    properties:
      datastore_type: mongodb
      datastore_version: 2.6.1
      instances:
        - flavor: m1.heat
          volume_size: 1
          networks:
            - port: port1
        - flavor: m1.heat
          volume_size: 1
          networks:
            - port: port2
        - flavor: m1.heat
          volume_size: 1
          networks:
            - port: port3
'''


class FakeTroveCluster(object):
    def __init__(self, status='ACTIVE'):
        self.name = 'cluster'
        self.id = '1189aa64-a471-4aa3-876a-9eb7d84089da'
        self.ip = ['10.0.0.1']
        self.instances = [
            {'id': '416b0b16-ba55-4302-bbd3-ff566032e1c1', 'status': status},
            {'id': '965ef811-7c1d-47fc-89f2-a89dfdd23ef2', 'status': status},
            {'id': '3642f41c-e8ad-4164-a089-3891bf7f2d2b', 'status': status}]

    def delete(self):
        pass


class FakeFlavor(object):
    def __init__(self, id, name):
        self.id = id
        self.name = name


class FakeVersion(object):
    def __init__(self, name="2.6.1"):
        self.name = name


class TroveClusterTest(common.HeatTestCase):

    def setUp(self):
        super(TroveClusterTest, self).setUp()

        self.tmpl = template_format.parse(stack_template)
        self.stack = utils.parse_stack(self.tmpl)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns['cluster']

        self.patcher_client = mock.patch.object(cluster.TroveCluster,
                                                'client')
        mock_client = self.patcher_client.start()
        self.client = mock_client.return_value
        self.troveclient = mock.Mock()
        self.troveclient.flavors.get.return_value = FakeFlavor(1, 'm1.heat')
        self.patchobject(neutron.NeutronClientPlugin,
                         'find_resourceid_by_name_or_id',
                         return_value='someportid')
        self.troveclient.datastore_versions.list.return_value = [FakeVersion()]
        self.patchobject(trove.TroveClientPlugin, 'client',
                         return_value=self.troveclient)

    def tearDown(self):
        super(TroveClusterTest, self).tearDown()
        self.patcher_client.stop()

    def _create_resource(self, name, snippet, stack):
        tc = cluster.TroveCluster(name, snippet, stack)
        self.client.clusters.create.return_value = FakeTroveCluster()
        self.client.clusters.get.return_value = FakeTroveCluster()
        scheduler.TaskRunner(tc.create)()
        return tc

    def test_create(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        expected_state = (tc.CREATE, tc.COMPLETE)
        self.assertEqual(expected_state, tc.state)
        args = self.client.clusters.create.call_args[1]
        self.assertEqual([{'flavorRef': '1', 'volume': {'size': 1},
                           'nics': [{'port-id': 'someportid'}]},
                          {'flavorRef': '1', 'volume': {'size': 1},
                           'nics': [{'port-id': 'someportid'}]},
                          {'flavorRef': '1', 'volume': {'size': 1},
                           'nics': [{'port-id': 'someportid'}]}],
                         args['instances'])
        self.assertEqual('mongodb', args['datastore'])
        self.assertEqual('2.6.1', args['datastore_version'])
        self.assertEqual('1189aa64-a471-4aa3-876a-9eb7d84089da',
                         tc.resource_id)
        self.assertEqual('clusters', tc.entity)

    def test_attributes(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        self.assertEqual(['10.0.0.1'], tc.FnGetAtt('ip'))
        self.assertEqual(['416b0b16-ba55-4302-bbd3-ff566032e1c1',
                          '965ef811-7c1d-47fc-89f2-a89dfdd23ef2',
                          '3642f41c-e8ad-4164-a089-3891bf7f2d2b'],
                         tc.FnGetAtt('instances'))

    def test_delete(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        fake_cluster = FakeTroveCluster()
        fake_cluster.task = {'name': 'NONE'}
        self.client.clusters.get.side_effect = [fake_cluster,
                                                fake_cluster,
                                                troveexc.NotFound()]
        scheduler.TaskRunner(tc.delete)()
        self.assertEqual((tc.DELETE, tc.COMPLETE), tc.state)

    def test_delete_not_found(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        self.client.clusters.get.side_effect = troveexc.NotFound()
        scheduler.TaskRunner(tc.delete)()
        self.assertEqual((tc.DELETE, tc.COMPLETE), tc.state)
        self.client.clusters.get.assert_called_with(tc.resource_id)
        self.assertEqual(2, self.client.clusters.get.call_count)

    def test_delete_incorrect_status(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        fake_cluster_bad = FakeTroveCluster()
        fake_cluster_bad.task = {'name': 'BUILDING'}
        fake_cluster_bad.delete = mock.Mock()
        fake_cluster_ok = FakeTroveCluster()
        fake_cluster_ok.task = {'name': 'NONE'}
        fake_cluster_ok.delete = mock.Mock()
        self.client.clusters.get.side_effect = [fake_cluster_bad,
                                                # two for cluster_delete method
                                                fake_cluster_bad,
                                                fake_cluster_ok,
                                                # for _refresh_cluster method
                                                troveexc.NotFound()]
        scheduler.TaskRunner(tc.delete)()
        self.assertEqual((tc.DELETE, tc.COMPLETE), tc.state)
        fake_cluster_bad.delete.assert_not_called()
        fake_cluster_ok.delete.assert_called_once_with()

    def test_delete_not_found_during_delete(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        fake_cluster = FakeTroveCluster()
        fake_cluster.task = {'name': 'NONE'}
        fake_cluster.delete = mock.Mock(side_effect=[troveexc.NotFound()])
        self.client.clusters.get.side_effect = [fake_cluster,
                                                fake_cluster,
                                                troveexc.NotFound()]
        scheduler.TaskRunner(tc.delete)()
        self.assertEqual((tc.DELETE, tc.COMPLETE), tc.state)
        self.assertEqual(1, fake_cluster.delete.call_count)

    def test_delete_already_deleting(self):
        tc = self._create_resource('cluster', self.rsrc_defn, self.stack)
        fake_cluster = FakeTroveCluster()
        fake_cluster.task = {'name': 'DELETING'}
        fake_cluster.delete = mock.Mock()
        self.client.clusters.get.side_effect = [fake_cluster,
                                                fake_cluster,
                                                troveexc.NotFound()]
        scheduler.TaskRunner(tc.delete)()
        self.assertEqual((tc.DELETE, tc.COMPLETE), tc.state)
        self.assertEqual(0, fake_cluster.delete.call_count)

    def test_validate_ok(self):
        tc = cluster.TroveCluster('cluster', self.rsrc_defn, self.stack)
        self.assertIsNone(tc.validate())

    def test_validate_invalid_dsversion(self):
        props = self.tmpl['resources']['cluster']['properties'].copy()
        props['datastore_version'] = '2.6.2'
        self.rsrc_defn = self.rsrc_defn.freeze(properties=props)
        tc = cluster.TroveCluster('cluster', self.rsrc_defn, self.stack)
        ex = self.assertRaises(exception.StackValidationFailed, tc.validate)
        error_msg = ('Datastore version 2.6.2 for datastore type mongodb is '
                     'not valid. Allowed versions are 2.6.1.')
        self.assertEqual(error_msg, str(ex))

    def test_validate_invalid_flavor(self):
        self.troveclient.flavors.get.side_effect = troveexc.NotFound()
        self.troveclient.flavors.find.side_effect = troveexc.NotFound()
        props = copy.deepcopy(self.tmpl['resources']['cluster']['properties'])
        props['instances'][0]['flavor'] = 'm1.small'
        self.rsrc_defn = self.rsrc_defn.freeze(properties=props)
        tc = cluster.TroveCluster('cluster', self.rsrc_defn, self.stack)
        ex = self.assertRaises(exception.StackValidationFailed, tc.validate)
        error_msg = ("Property error: "
                     "resources.cluster.properties.instances[0].flavor: "
                     "Error validating value 'm1.small': Not Found (HTTP 404)")
        self.assertEqual(error_msg, str(ex))

    def test_validate_no_net(self):
        props = self.tmpl['resources']['cluster']['properties'].copy()
        props['instances'][0]['networks'][0].pop('port')
        self.rsrc_defn = self.rsrc_defn.freeze(properties=props)
        tc = cluster.TroveCluster('cluster', self.rsrc_defn, self.stack)
        self.assertRaises(exception.StackValidationFailed, tc.validate)

    def test_validate_both_net_and_port_used(self):
        props = self.tmpl['resources']['cluster']['properties'].copy()
        props['instances'][0]['networks'][0]['network'] = 'mynetwork'
        self.rsrc_defn = self.rsrc_defn.freeze(properties=props)
        tc = cluster.TroveCluster('cluster', self.rsrc_defn, self.stack)
        self.assertRaises(exception.ResourcePropertyConflict, tc.validate)
